"""
Pollenisator client GUI window.
"""
import threading
import traceback
import tkinter.filedialog
import tkinter as tk
import tkinter.messagebox
import tkinter.simpledialog
import tkinter.ttk as ttk
import uuid
from customtkinter import *
import sys
import os
from tkinter import TclError
import json
import re
from PIL import ImageTk, Image
import importlib
import pkgutil
import socketio
from pollenisatorgui.core.application.dialogs.ChildDialogToolsInstalled import ChildDialogToolsInstalled
from pollenisatorgui.core.application.scrollableframexplateform import ScrollableFrameXPlateform
from pollenisatorgui.core.application.terminalswidget import TerminalsWidget
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.components.tag import TagInfos
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.application.treeviews.PentestTreeview import PentestTreeview
from pollenisatorgui.core.application.treeviews.CommandsTreeview import CommandsTreeview
from pollenisatorgui.core.application.dialogs.ChildDialogCombo import ChildDialogCombo
from pollenisatorgui.core.application.dialogs.ChildDialogQuestion import ChildDialogQuestion
from pollenisatorgui.core.application.dialogs.ChildDialogConnect import ChildDialogConnect
from pollenisatorgui.core.application.dialogs.ChildDialogPentests import ChildDialogPentests
from pollenisatorgui.core.application.dialogs.ChildDialogException import ChildDialogException
from pollenisatorgui.core.application.dialogs.ChildDialogFileParser import ChildDialogFileParser
from pollenisatorgui.core.application.dialogs.ChildDialogEditPassword import ChildDialogEditPassword
from pollenisatorgui.core.application.statusbar import StatusBar
from pollenisatorgui.core.components.apiclient import APIClient, ErrorHTTP
from pollenisatorgui.core.components.scanmanager import ScanManager
from pollenisatorgui.core.components.admin import AdminView
from pollenisatorgui.core.components.scriptmanager import ScriptManager
from pollenisatorgui.core.components.settings import Settings
from pollenisatorgui.core.components.filter import Filter
from pollenisatorgui.core.controllers.toolcontroller import ToolController
from pollenisatorgui.core.forms.formpanel import FormPanel
from pollenisatorgui.core.models.port import Port
from pollenisatorgui.core.views.checkinstanceview import CheckInstanceView
from pollenisatorgui.core.views.checkitemview import CheckItemView
from pollenisatorgui.core.views.ipview import IpView
import pollenisatorgui.modules
import customtkinter
import tkinterDnD
from ttkwidgets import tooltips
from pollenisatorgui.core.application.pollenisatorentry import PopoEntry
from pollenisatorgui.core.components.logger_config import logger
from pollenisatorgui.modules.module import Module

class FloatingHelpWindow(CTkToplevel):
    """floating basic window with helping text inside
    Inherit tkinter TopLevel
    Found on the internet (stackoverflow) but did not keep link sorry...
    """

    def __init__(self, w, h, posx, posy, *args, **kwargs):
        CTkToplevel.__init__(self, *args, **kwargs)
        self.attributes("-type", "dialog")
        self.title('Help: search')
        self.x = posx
        self.y = posy
        self.geometry(str(w)+"x"+str(h)+"+"+str(posx)+"+"+str(posy))
        self.resizable(0, 0)
        self.configure(bg='light yellow')
        self.grip = tk.Label(self, bitmap="gray25")
        self.grip.pack(side="left", fill="y")
        label = tk.Label(self, bg='light yellow', fg='black',
                         justify=tk.LEFT, text=Filter.help())
        label.pack()
        self.overrideredirect(True)
        self.grip.bind("<ButtonPress-1>", self.startMove)
        self.grip.bind("<ButtonRelease-1>", self.stopMove)
        self.grip.bind("<B1-Motion>", self.onMotion)

    def startMove(self, event):
        """ Floating window dragging started
            Args:
                event: event.x and event.y hold the new position of the window
        """
        self.x = event.x
        self.y = event.y

    def stopMove(self, _event=None):
        """ Floating window dragging stopped
            Args:
                _event: Not used but mandatory
        """
        self.x = None
        self.y = None

    def onMotion(self, event):
        """ Floating window dragging ongoing
            Args:
                event: event.x and event.y hold the new position of the window
        """
        deltax = event.x - self.x
        deltay = event.y - self.y
        x = self.winfo_x() + deltax
        y = self.winfo_y() + deltay
        self.geometry("+%s+%s" % (x, y))


class AutocompleteEntry(PopoEntry):
    """Inherit PopoEntry.
    An entry with an autocompletion ability.
    Found on the internet : http://code.activestate.com/recipes/578253-an-entry-with-autocompletion-for-the-tkinter-gui/
    But a bit modified.
    """

    def __init__(self, settings, *args, **kwargs):
        """Constructor
        Args:
            settings: a dict of Settings:
                * histo_filters: number of history search to display
            args: not used
            kwargs: 
                * width: default to 100
        """
        PopoEntry.__init__(self, *args, **kwargs)
        self.width = kwargs.get("width",100)
        self.lista = set()
        self.var = self.cget("textvariable")
        if self.var is None:
            self.var = tk.StringVar()
            self.configure(textvariable=self.var)
        self.var.trace('w', self.changed)
        self.bind("<Tab>", self.selection)
        self.bind("<Right>", self.selection)
        self.bind("<Up>", self.upArrow)
        self.bind("<Down>", self.downArrow)
        self.settings = settings
        self.server_time = None
        self.lb = None
        self.lb_up = False
        

    def changed(self, _name=None, _index=None, _mode=None):
        """
        Called when the entry is modified. Perform autocompletion.
        Args:
            _name: not used but mandatory for tk.StringVar.trace
            _index: not used but mandatory for tk.StringVar.trace
            _mode: not used but mandatory for tk.StringVar.trace
        """
        words = self.comparison()
        if words:
            if not self.lb_up:
                self.lb = tk.Listbox(width=self.width, fg=utils.getTextColor(), bg=utils.getBackgroundColor())
                self.lb.bind("<Double-Button-1>", self.selection)
                self.lb.bind("<Right>", self.selection)
                self.lb.bind("<Leave>", self.quit)
                self.bind("<Escape>", self.quit)
                self.lb.place(x=self.winfo_x()+133,
                                y=self.winfo_y()+self.winfo_height()+20)
                self.lb_up = True
            self.lb.delete(0, tk.END)
            for w in words:
                self.lb.insert(tk.END, w)
        else:
            self.quit()

    def quit(self, _event=None):
        """
        Callback function to destroy the label shown
        Args:
            _event: not used but mandatory
        """
        if self.lb_up:
            self.lb.destroy()
            self.lb_up = False
    
    def reset(self):
        """
        quit and reset filter bar
        """
        self.quit()
        self.var.set("")

    def selection(self, _event=None):
        """
        Called when an autocompletion option is chosen. 
        Change entry content and close autocomplete.
        Args:
            _event: not used but mandatory
        """
        if self.lb_up:
            self.var.set(self.lb.get(tk.ACTIVE))
            self.lb.destroy()
            self.lb_up = False
            self.icursor(tk.END)
            self.focus_set()
            self.focus()
            #self.changed()
            return 'break'

    def upArrow(self, _event=None):
        """
        Called when the up arrow is pressed. Navigate in autocompletion options
        Args:
            _event: not used but mandatory
        """
        if self.lb_up:
            if self.lb.curselection() == ():
                index = '0'
            else:
                index = self.lb.curselection()[0]
            if index != '0':
                self.lb.selection_clear(first=index)
                index = str(int(index)-1)
                self.lb.selection_set(first=index)
                self.lb.activate(index)

    def downArrow(self, _event=None):
        """
        Called when the down arrow is pressed. Navigate in autocompletion options
        Args:
            _event: not used but mandatory
        """
        if self.lb_up:
            if self.lb.curselection() == ():
                index = '0'
            else:
                index = self.lb.curselection()[0]
            if index != tk.END:
                self.lb.selection_clear(first=index)
                index = str(int(index)+1)
                self.lb.selection_set(first=index)
                self.lb.activate(index)

    def comparison(self):
        """
        Search suggestions in regard of what is in the entry
        """
        values = set(self.settings.local_settings.get("histo_filters", []))
        self.lista = values
        content = self.var.get().strip()
        if content == "":
            return []
        pattern = re.compile('.*' + re.escape(content) + '.*')
        return [w for w in self.lista if re.match(pattern, w)]

def iter_namespace(ns_pkg):
    # Specifying the second argument (prefix) to iter_modules makes the
    # returned name an absolute name instead of a relative one. This allows
    # import_module to work without having to do additional modification to
    # the name.
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")

class ButtonNotebook(CTkFrame):
    def __init__(self, parent, callbackSwitch, closeCallbackSwitch):
        super().__init__(parent)
        self.frameButtons = CTkFrame(self, fg_color=utils.getStrongColor())
        self.callbackSwitch = callbackSwitch
        self.closeCallbackSwitch = closeCallbackSwitch
        self.tabs = {}
        self.current = None
        self.frameButtons.pack(side="left", anchor="nw", fill=tk.Y)
        self.btns = {}
        self.lbl = None

    def add(self, widget, name, order, image):
        if name not in self.tabs:
            self.tabs[name] = {"widget":widget, "image":image, "order": order, "name":name}
            widget.pack_forget()
            btn = CTkButton(self.frameButtons, text=name, image=image,  fg_color=utils.getStrongColor() , hover_color=utils.getStrongActiveColor(), compound=tk.TOP)
            self.btns[name] = btn
            btn.bind("<Button-1>", self.clicked)
            self.redraw()
            

    def redraw(self):
        for btn in self.btns.values():
            btn.pack_forget()
        if self.lbl:
            self.lbl.pack_forget()
        btns = sorted(self.tabs.values(), key=lambda x:x["order"])
        for btn in btns:
            
            self.btns[btn["name"]].pack(side="top", fill=tk.X, anchor="nw")
        self.image = Image.open(utils.getIcon("LogoPollenisator.png"))
        img = CTkImage(light_image=self.image, dark_image=self.image, size=(100, 123))
        self.lbl = CTkLabel(self.frameButtons, text="", image=img)
        self.lbl.pack(side="bottom", fill=tk.X, pady=5,anchor="sw")

    def clicked(self, event):
        widget = event.widget.master
        name = widget.cget("text")
        self.select(name)

    def getOpenTabName(self):
        return self.current


    def delete(self, name):
        if name in self.tabs:
            del self.tabs[name]
            self.btns[name].pack_forget()

    def select(self, name):
        if self.current == name:
            return
        if self.current:
            self.tabs[self.current]["widget"].pack_forget()
            self.btns[self.current].configure(fg_color=utils.getStrongColor())

            self.closeCallbackSwitch(self.current, name)
        self.current = name
        self.btns[name].configure(fg_color=utils.getStrongActiveColor())
        self.tabs[name]["widget"].pack(side="right", expand=True, anchor="center", fill=tk.BOTH)
        self.callbackSwitch(name)


class Appli(customtkinter.CTk, tkinterDnD.tk.DnDWrapper):#HACK to make work tkdnd with CTk
    """
    Main tkinter graphical application object.
    """
    version_compatible = "2.7.*"

    
    def _init_tkdnd(master: tk.Tk) -> None: #HACK to make work tkdnd with CTk
        """Add the tkdnd package to the auto_path, and import it"""
        #HACK Copied from directory with a package_dir updated
        platform = master.tk.call("tk", "windowingsystem")

        if platform == "win32":
            folder = "windows"
        elif platform == "x11":
            folder = "linux"
        elif platform == "aqua":
            folder = "mac"
        package_dir = os.path.join(os.path.dirname(os.path.abspath(tkinterDnD.tk.__file__)), folder)
        master.tk.call('lappend', 'auto_path', package_dir)
        TkDnDVersion = master.tk.call('package', 'require', 'tkdnd')
        return TkDnDVersion

    def __init__(self):
        """
        Initialise the application

        """
        # Lexic:
        # view frame : the frame in the tab that will hold forms.
        # Tree view : the tree on the left of the window.
        # frame tree view : a frame around the tree view (useful to attach a scrollbar to a treeview)
        # paned : a Paned widget is used to separate two other widgets and display a one over the other if desired
        #           Used to separate the treeview frame and view frame.
        super().__init__()
        self.TkDnDVersion = self._init_tkdnd()  #HACK to make work tkdnd with CTk
        self.quitting = False
        self.settingViewFrame = None
        self.scanManager = None  #  Loaded when clicking on it if linux only
        self.scanViewFrame = None
        self.admin = None
        self.nbk = None
        self.notif_handlers = []
        self.sio = None #socketio client
        self.initialized = False
        self.settings = Settings()
        self.notif_processing_timer = None

        utils.setStyle(self, self.settings.local_settings.get("dark_mode", False))
        self.main_tab_img = CTkImage(
            Image.open(utils.getIcon("tab_main.png")), size=(30, 30))
        self.commands_tab_img = CTkImage(
            Image.open(utils.getIcon("tab_commands.png")), size=(30, 30))
        self.scan_tab_img = CTkImage(
            Image.open(utils.getIcon("tab_scan.png")), size=(30, 30))
        self.settings_tab_img = CTkImage(
            Image.open(utils.getIcon("tab_settings.png")), size=(30, 30))
        self.admin_tab_img = CTkImage(
            Image.open(utils.getIcon("tab_admin.png")), size=(30, 30))
        # HISTORY : Main view and command where historically in the same view;
        # This results in lots of widget here with a confusing naming style
        #### core components (Tab menu on the left objects)####
        #### MAIN VIEW ####
        self.mainPageFrame = None
        self.paned = None
        self.proxyFrameMain = None
        self.viewframe = None
        self.frameTw = None
        self.treevw = None
        self.datamanager = None
        self.terminals = None
        self.myscrollbarMain = None
        #### COMMAND VIEW ####
        self.commandsPageFrame = None
        self.commandPaned = None
        self.commandsFrameTw = None
        self.commandsViewFrame = None
        self.myscrollbarCommand = None
        self.commandsTreevw = None
        #### SEARCH BAR ####
        # boolean set to true when the main tree view is displaying search results
        self.searchMode = False
        self.searchBar = None  # the search bar component
        self.btnHelp = None  # help button on the right of the search bar
        self.photo = None  # the ? image
        self.helpFrame = None  # the floating help frame poping when the button is pressed
        dir_path = utils.getIcon("favicon.png")
        img = tk.PhotoImage(file=dir_path)
        self.resizable(True, True)
        self.iconphoto(True, img)
        self.minsize(width=400, height=400)
        self.resizable(True, True)
        self.title("Pollenisator")
        monitor = utils.get_screen_where_widget(self)
        self.geometry(f"{monitor.width}x{monitor.height}+{monitor.x}+{monitor.y}")
        self.protocol("WM_DELETE_WINDOW", self.onClosing)
        self.datamanager = DataManager.getInstance()
        self.initModules()
        apiclient = APIClient.getInstance()
        self.topviewframe = None
        self.scanManager = ScanManager(self, self.nbk, self.treevw, apiclient.getCurrentPentest(), self.settings)
        apiclient.appli = self
        opened, errored = self.openConnectionDialog()
        if errored:
            return
        if not opened:
            try:
                self.wait_visibility(self.tk)
            except tk.TclError: #closed dialog
                return
            except AttributeError:  # closed and destroyed
                return
            opened, errored = self.openConnectionDialog(force=True)
            if errored:
                return
            self.openPentestsWindow()
        self.loadModulesInfos() 
        self.scanManager.nbk = self.nbk #FIXME ORDER, INITIALISATION of SCAN MANAGERis too early
        self.scanManager.linkTw = self.treevw


    def start_autoscan(self):
        return self.scanManager.startAutoscan()

    def stop_autoscan(self):
        return self.scanManager.stop()

    # OVERRIDE tk.Tk.report_callback_exception
    def report_callback_exception(self, exc, val, tb):
        self.show_error(exc, val, tb)
        
    def quit(self):
        super().quit()
        self.quitting = True
        return

    def forceUpdate(self, api_version, my_version):
        tkinter.messagebox.showwarning("Update necessary", f"Clash of version. Expecting API {api_version} and you are compatible with {my_version}. Please reinstall following the instructions in README. (git pull; pip install .)")

    def openConnectionDialog(self, force=False):
        # Connect to database and choose database to open
        apiclient = APIClient.getInstance()
        abandon = False
        connectionTest = apiclient.tryConnection(force=force)
        if force:
            apiclient.disconnect()
        res = apiclient.tryAuth()
        if not res: 
            apiclient.disconnect()
        while (not connectionTest or not apiclient.isConnected()) and not abandon:
            abandon = self.promptForConnection() is None
            connectionTest = apiclient.tryConnection(force=force)
        if not abandon:
            apiclient = APIClient.getInstance()
            apiclient.attach(self)
            srv_version = apiclient.getVersion()
            if int(Appli.version_compatible.split(".")[0]) != int(srv_version.split(".")[0]):
                self.forceUpdate(srv_version, Appli.version_compatible)
                self.onClosing()
                return False, True
            if int(Appli.version_compatible.split(".")[1]) != int(srv_version.split(".")[1]):
                self.forceUpdate(srv_version, Appli.version_compatible)
                self.onClosing()
                return False, True
            if self.sio is not None:
                self.sio.disconnect()
            self.sio = socketio.Client()
            @self.sio.event
            def notif(data):
                self.handleNotif(data)
            
            @self.sio.event
            def test(data):
                tk.messagebox.showinfo("test", "test socket working received data : "+str(data))
           # Event handler for client disconnection
            @self.sio.event
            def disconnect():
                logger.debug(f"Client disconnected")
                t = threading.Timer(0.5, self.reconnect)
                t.start()
                return True
            try:
                self.sio.connect(apiclient.api_url)
            except RuntimeError:
                # tryto reconnect
                self.sio.connect(apiclient.api_url)
            pentests = apiclient.getPentestList()
            self._initMenuBar()

            if pentests is None:
                pentests = []
            else:
                pentests_names = [x["nom"] for x in pentests][::-1]
            if apiclient.getCurrentPentest() != "" and apiclient.getCurrentPentest() in pentests_names:
                self.openPentest(apiclient.getCurrentPentest())
            else:
                ret = self.openPentestsWindow(pentests=pentests)
                if ret is None:
                    pass
                    # self.onClosing()
                    # try:
                    #     self.destroy()
                    # except tk.TclError:
                    #     pass   
                    # return False, False
            self.initialized = True

        else:
            self.onClosing()
            try:
                self.destroy()
            except tk.TclError:
                pass
        return apiclient.isConnected(), False
    
    def reconnect(self):
        apiclient = APIClient.getInstance()
        if not apiclient.isConnected() or self.quitting:
            return
        logger.debug("SocketIO reconnection quiitng="+str(self.quitting))
        try:
            self.sio.connect(apiclient.api_url)
        except ConnectionError:
            logger.debug("SocketIO reconnection failed")
            t = threading.Timer(0.5, self.reconnect)
            t.start()
            return
        self.sio.emit("registerForNotifications", {"token":apiclient.getToken(), "pentest":apiclient.getCurrentPentest()})

    def initModules(self):
        discovered_plugins = {
            name: importlib.import_module(name)
            for finder, name, ispkg
            in iter_namespace(pollenisatorgui.modules) if not name.endswith(".Module")
        }
        self.modules = []
        from pollenisatorgui.modules.module import REGISTRY
        for name, module_class in REGISTRY.items():
            if name != "Module":
                module_obj = module_class(self, self.settings, self)
                self.modules.append({"name": module_obj.tabName, "object":module_obj, "view":None, "img":CTkImage(Image.open(utils.getIcon(module_obj.iconName)), size=(30, 30))})
        
    def loadModulesInfos(self):
        for module in self.modules:
            if callable(getattr(module["object"], "loadModuleInfo", False)):
                module["object"].loadModuleInfo()

    def show_error(self, *args):
        """Callback for tk.Tk.report_callback_exception.
        Open a window to display exception with some actions possible

        Args:
            args: 3 args are required for tk.Tk.report_callback_exception event to be given to traceback.format_exception(args[0], args[1], args[2])
        
        Raises:
            If an exception occurs in this handler thread, will print it and exit with exit code 1
        """
        try:
            if self.quitting:
                return
            err = traceback.format_exception(args[0], args[1], args[2])
            err = "\n".join(err)
            
            if args[0] is ErrorHTTP: # args[0] is class of ecx
                if args[1].response.status_code == 401:
                    tk.messagebox.showerror("Disconnected", "You are not connected.")
                    self.openConnectionDialog(force=True)
                    return
            dialog = ChildDialogException(self, 'Exception occured', err)
            apiclient = APIClient.getInstance()
            apiclient.reportError(err)
            try:
                self.wait_window(dialog.app)
            except tk.TclError:
                sys.exit(1)
        except Exception as e:
            print("Exception in error handler "+str(e))
            sys.exit(1)
        
        
    def promptForConnection(self):
        """Close current database connection and open connection form for the user
        
        Returns: 
            The number of pollenisator database found, 0 if the connection failed."""
        apiclient = APIClient.getInstance()
        apiclient.reinitConnection()
        connectDialog = ChildDialogConnect(self)
        self.wait_window(connectDialog.app)
        return connectDialog.rvalue

    def changeMyPassword(self):
        """Allows the current user to change its password"""
        apiclient = APIClient.getInstance()
        connected_user = apiclient.getUser()
        if connected_user is None:
            tk.messagebox.showerror("Change password", "You are not connected")
            return 
        dialog = ChildDialogEditPassword(self, connected_user)
        self.wait_window(dialog.app)
        
    def disconnect(self):
        """Remove the session cookie"""
        APIClient.getInstance().disconnect()
        
        self.openConnectionDialog(force=True)

    def submitIssue(self):
        """Open git issues in browser"""
        import webbrowser
        webbrowser.open_new_tab("https://github.com/AlgoSecure/Pollenisator/issues")

    def notify(self, notification):
        for notif_handler in self.notif_handlers:
            if notif_handler["pentest"] is not None and notif_handler["pentest"] != notification["db"]:
                continue
            if notif_handler["collection"] is not None and notif_handler["collection"] != notification["collection"]:
                continue
            if notif_handler["iid"] is not None and notif_handler["iid"] != str(notification["iid"]):
                continue
            if notif_handler["notif_name"] is not None and notif_handler["notif_name"] != notification["action"]:
                continue
            
            notif_handler["handler"](notification)
    
    def subscribe_notification(self, notif_name, handler, pentest=None, collection=None, iid=None):
        if handler is None:
            return
        self.notif_handlers.append({"pentest":pentest, "collection":collection, "iid":iid, "notif_name":notif_name, "handler":handler})
    
    def unsubscribe_notification(self, notif_name, pentest=None, collection=None, iid=None):
        i = 0
        while self.notif_handlers and i < len(self.notif_handlers):
            notif_handler = self.notif_handlers[i]
            if notif_handler["pentest"] == pentest and notif_handler["notif_name"] == notif_name \
                and notif_handler["collection"] == collection and str(notif_handler["iid"]) == str(iid): 
                del self.notif_handlers[i]
            else:
                i+=1

    def handleNotif(self, notification):
        notification = json.loads(notification, cls=utils.JSONDecoder)
        self.notify(notification)
        self.datamanager.handleNotification(notification)
        # self.notif_waiting.append(notification)
        # if self.notif_processing_timer is None and not self.quitting:
        #     self.notif_processing_timer = threading.Timer(1, self.notif_processing)
        #     self.notif_processing_timer.start()


    # def notif_processing(self):
    #     notif_treatments = deepcopy(self.notif_waiting)
    #     self.notif_waiting = []
    #     notif_treatments.sort(key=lambda x: x["time"])
    #     while notif_treatments:
    #         notification = notif_treatments.pop(0)
    #         self.datamanager.handleNotification(notification)
    #         self.notify(notification)
    #     self.notif_processing_timer = None
        
    def onClosing(self):
        """
        Close the application properly.
        """
        if self.quitting:
            return
        self.quitting = True
        self.closePentest()
        print("Stopping application...")
        if self.sio is not None:
            self.sio.disconnect()
            self.sio.eio.disconnect()
        self.quit()

    def reopen(self, event=None):
        self.treevw.reopen()

    def _initMenuBar(self):
        """
        Create the bar menu on top of the screen.
        """
        menubar = utils.craftMenuWithStyle(self)
        self.configure(menu=menubar)

        self.bind('<F5>', self.refreshView)
        self.bind('<F6>', self.reopen)
        self.bind('<Control-o>', self.openPentestsWindow)
        fileMenu =  utils.craftMenuWithStyle(menubar)
        fileMenu.add_command(label="Pentests management (Ctrl+o)",
                             command=self.openPentestsWindow)
        fileMenu.add_command(label="Connect to server", command=self.promptForConnection)
        fileMenu.add_command(label="Export commands",
                             command=self.exportCommands)
        fileMenu.add_command(label="Import commands",
                             command=self.importCommands)
        fileMenu.add_command(label="Export cheatsheet",
                             command=self.exportCheatsheet)
        fileMenu.add_command(label="Import cheatsheet",
                             command=self.importCheatsheet)
        fileMenu.add_command(label="Import defect templates",
                             command=self.importDefectTemplates)                     

        fileMenu.add_command(label="Exit", command=self.onClosing)
        fileMenu2 = utils.craftMenuWithStyle(menubar)
        fileMenu2.add_command(label="Import existing tools results ...",
                              command=self.importExistingTools)
        fileMenu2.add_command(label="Reset unfinished tools",
                              command=self.resetUnfinishedTools)
        fileMenu2.add_command(label="Test local tools",
                              command=self.wrapperTestLocalTools)
        fileMenu2.add_command(label="Refresh (F5)",
                              command=self.refreshView)
        fileMenuUser = utils.craftMenuWithStyle(menubar)
        fileMenuUser.add_command(label="Change your password",
                              command=self.changeMyPassword)
        fileMenuUser.add_command(label="Disconnect", command=self.disconnect)
        fileMenu3 = utils.craftMenuWithStyle(menubar)
        fileMenu3.add_command(label="Submit a bug or feature",
                              command=self.submitIssue)
        fileMenuDebug = utils.craftMenuWithStyle(menubar)
        fileMenuDebug.add_command(label="Socket test", command=self.socketTest)
        menubar.add_cascade(label="Database", menu=fileMenu)
        menubar.add_cascade(label="Scans", menu=fileMenu2)
        menubar.add_command(label="Scripts...", command=self.openScriptModule)
        menubar.add_cascade(label="User", menu=fileMenuUser)
        menubar.add_cascade(label="Help", menu=fileMenu3)
        menubar.add_cascade(label="Debug", menu=fileMenuDebug)
        
    def socketTest(self):
        print("EMIT TEST")
        self.sio.emit("test", {"pentest": APIClient.getInstance().getCurrentPentest()})
        print("TEST SENT, WAITING FOR RESPONSE")

    


    def initMainView(self):
        """
        Fill the main view tab menu
        """
        self.mainPageFrame = CTkFrame(self.topviewframe)
        searchFrame = CTkFrame(self.mainPageFrame, fg_color=utils.getBackgroundSecondColor())
        filterbar_frame = CTkFrame(searchFrame, fg_color="transparent")
        self.image_filter = CTkImage(Image.open(utils.getIcon("filter.png")))
        lblSearch = CTkLabel(filterbar_frame, text="Filter bar", image=self.image_filter, compound = "left")
        lblSearch.pack(side="left", fill=tk.NONE)
        self.searchBar = AutocompleteEntry(self.settings, filterbar_frame)
        #self.searchBar = PopoEntry(filterbar_frame, width=108)
        self.searchBar.bind('<Return>', self.newSearch)
        self.searchBar.bind('<KP_Enter>', self.newSearch)
        # searchBar.bind("<Button-3>", self.do_popup)
        self.searchBar.pack(side="left", fill="x", expand=True)
        self.textSearchVal = tk.BooleanVar()
        self.textSearchVal.set(self.settings.local_settings.get("textsearch", False))
        
        checkbox_text_search = CTkSwitch(filterbar_frame, text="Text search", variable=self.textSearchVal, command=self.textSearchChanged)
        checkbox_text_search.pack(side="left", padx=5)
        self.keep_parents_val = tk.BooleanVar()
        self.keep_parents_val.set(self.settings.local_settings.get("keep_parents", True))
        checkbox_keep_parent = CTkSwitch(filterbar_frame, text="Keep parents", variable=self.keep_parents_val, command=self.keepParentsChanged)
        checkbox_keep_parent.pack(side="left", padx=5)
        self.search_icon = tk.PhotoImage(file=utils.getIcon("search.png"))
        btnSearchBar = ttk.Button(filterbar_frame, text="", image=self.search_icon, style="iconbis.TButton", tooltip="Filter elements based of complex query or only text if textsearch is selected", width=10, command=self.newSearch)
        btnSearchBar.pack(side="left", fill="x")
        image=Image.open(utils.getIcon("reset_small.png"))
        self.reset_icon = ImageTk.PhotoImage(image)
        btnReset = ttk.Button(filterbar_frame, image=self.reset_icon, text="",  style="iconbis.TButton", tooltip="Reset search bar filter", width=10, command=self.endSearch)
        btnReset.pack(side="left", fill="x")
        self.photo = CTkImage(Image.open(utils.getHelpIconPath()))
        self.helpFrame = None
        self.btnHelp = CTkButton(filterbar_frame, text="",image=self.photo,  fg_color="transparent", width=10, command=self.showSearchHelp)

        self.btnHelp.pack(side="left")
        filterbar_frame.pack(side=tk.TOP,fill=tk.X)
        self.statusbar = StatusBar(searchFrame, self.statusbarClicked)
        self.statusbar.pack(side=tk.BOTTOM, fill=tk.X)
        searchFrame.pack(side="top", fill="x")
        #PANED PART
        self.paned = tk.PanedWindow(self.mainPageFrame, orient="horizontal")
        #RIGHT PANE : Canvas + frame
        
        self.proxyFrameMain = CTkFrame(self.paned)
        self.proxyFrameMain.rowconfigure(0, weight=1) 
        self.proxyFrameMain.columnconfigure(0, weight=1) 
        self.viewframe = ScrollableFrameXPlateform(self.proxyFrameMain)
        
        
        #LEFT PANE : Treeview
        self.left_pane = CTkFrame(self.paned)
        self.frameTw = CTkFrame(self.left_pane)
        self.frameTw.rowconfigure(1, weight=1) # Weight 1 sur un layout grid, sans ça le composant ne changera pas de taille en cas de resize
        self.frameTw.columnconfigure(0, weight=1) # Weight 1 sur un layout grid, sans ça le composant ne changera pas de taille en cas de resize
        self.treevw = PentestTreeview(self, self.frameTw)
        frameContext = CTkFrame(self.frameTw)
        self.btn_context = CTkSegmentedButton(frameContext, values=["Hosts", "Checklist"], width=200, height=45, command=self.checklistViewSwap)
        self.btn_context.set("Check" if self.settings.is_checklist_view() else "Hosts")
        self.btn_context.pack(side="left")
        frameContext.grid(row=0, column=0)
        self.treevw.initUI()
        self.scbVSel = CTkScrollbar(self.frameTw,
                                orientation=tk.VERTICAL,
                                command=self.treevw.yview)
        self.treevw.configure(yscrollcommand=self.scbVSel.set)
        self.treevw.grid(row=1, column=0, sticky=tk.NSEW)
        self.scbVSel.grid(row=1, column=1, sticky=tk.NS)
        # FILTER PANE:
        # self.filtersFrame = CTkFrame(self.left_pane)
        # self.initFiltersFrame(self.filtersFrame)
        # self.filtersFrame.pack(side="bottom", fill="x")
        # END PANE PREP
        self.frameTw.pack(side="top", fill=tk.BOTH, expand=True)
        
        self.left_pane.pack(side="left", fill=tk.BOTH, expand=True)
        self.paned.add(self.left_pane)
        self.proxyFrameMain.pack(side="right", fill=tk.BOTH, expand=True)
        self.viewframe.pack(fill=tk.BOTH, expand=1)
        
        self.paned.add(self.proxyFrameMain)

        self.paned.pack(fill=tk.BOTH, expand=1)
        self.mainPageFrame.pack(fill="both", expand=True)
        self.nbk.add(self.mainPageFrame, "Main View", order=Module.HIGH_PRIORITY, image=self.main_tab_img)
        
        self.nbk
    def searchbarSelectAll(self, _event=None):
        """
        Callback to select all the text in searchbar
        Args:
            _event: not used but mandatory
        """
        self.searchBar.select_range(0, 'end')
        self.searchBar.icursor('end')
        return "break"

   
    def checklistViewSwap(self, _event=None):
        is_checklist_view = "Check" in self.btn_context.get()
        settings = Settings()
        settings.local_settings["checklist_view"] = is_checklist_view
        settings.saveLocalSettings()
        self.treevw.checklistViewSwap()

    def initCommandsView(self):
        """Populate the command tab menu view frame with cool widgets"""
        self.commandsPageFrame = CTkFrame(self.topviewframe)
        self.commandPaned = tk.PanedWindow(self.commandsPageFrame, height=800)
        self.commandsFrameTw = CTkFrame(self.commandPaned)
        self.proxyFrameCommand = CTkFrame(self.commandPaned)
        self.commandsFrameTw.pack(expand=True)
        self.commandsViewFrame = ScrollableFrameXPlateform(self.proxyFrameCommand)
        self.commandsTreevw = CommandsTreeview(self, self.commandsFrameTw)
        scbVSel = CTkScrollbar(self.commandsFrameTw,
                                orientation=tk.VERTICAL,
                                command=self.commandsTreevw.yview)
        self.commandsTreevw.configure(yscrollcommand=scbVSel.set)
        self.commandsTreevw.grid(row=0, column=0, sticky=tk.NSEW)
        scbVSel.grid(row=0, column=1, sticky=tk.NS)
        self.commandPaned.add(self.commandsFrameTw)
        self.commandsViewFrame.pack(fill=tk.BOTH, expand=1)
        self.commandPaned.add(self.proxyFrameCommand)
        self.commandPaned.pack(fill=tk.BOTH, expand=1)
        self.commandsFrameTw.rowconfigure(0, weight=1) # Weight 1 sur un layout grid, sans ça le composant ne changera pas de taille en cas de resize
        self.commandsFrameTw.columnconfigure(0, weight=1) # Weight 1 sur un layout grid, sans ça le composant ne changera pas de taille en cas de resize
        self.nbk.add(self.commandsPageFrame, "Commands", order=Module.LOW_PRIORITY, image=self.commands_tab_img)


    def showSearchHelp(self, _event=None):
        """Called when the searchbar help button is clicked. Display a floating help window with examples
        Args:
            _event: not used but mandatory
        """
        if self.helpFrame is None:
            x, y = self.btnHelp.winfo_rootx(), self.btnHelp.winfo_rooty()
            self.helpFrame = FloatingHelpWindow(410, 400, x-380, y+40, self)
        else:
            self.helpFrame.destroy()
            self.helpFrame = None

    def beforeTabSwitch(self, current, new):
        if current is None:
            return
        for module in self.modules:
            if current.strip().lower() == module["name"].strip().lower():
                if hasattr(module["object"], "close"):
                    module["object"].close()

    def tabSwitch(self, tabName):
        """Called when the user click on the tab menu to switch tab. Add a behaviour before the tab switches.
        Args:
            tabName: the opened tab
        """
        apiclient = APIClient.getInstance()
        self.searchBar.quit()
        if tabName == "Main View":
            self.refreshUI()
        if tabName == "Commands":
            self.commandsTreevw.initUI()
        if apiclient.getCurrentPentest() is None or apiclient.getCurrentPentest() == "":
            opened = self.openPentestsWindow()
            if opened is None:
                return
        if tabName == "Scan":
            if apiclient.getCurrentPentest() != "":
                self.scanManager.refreshUI()
        elif tabName == "Settings":
            self.settings.reloadUI()
        elif tabName == "Admin":
            self.admin.refreshUI()
        else:
            for module in self.modules:
                if tabName.strip().lower() == module["name"].strip().lower():
                    module["object"].open(module["view"], self.topviewframe, self.treevw)

    def initSettingsView(self):
        """Add the settings view frame to the notebook widget and initialize its UI."""
        self.settingViewFrame = CTkFrame(self.topviewframe)
        self.settings.initUI(self.settingViewFrame)
        self.nbk.add(self.settingViewFrame, "Settings", order=Module.LAST_PRIORITY, image=self.settings_tab_img)

    def initScanView(self):
        """Add the scan view frame to the notebook widget. This does not initialize it as it needs a database to be opened."""
        self.scanViewFrame = CTkFrame(self.topviewframe)
        self.scanManager.initUI(self.scanViewFrame)
        self.nbk.add(self.scanViewFrame, "Scan", order=Module.HIGH_PRIORITY, image=self.scan_tab_img)

    def initAdminView(self):
        """Add the admin button to the notebook"""
        self.admin = AdminView(self.topviewframe)
        self.adminViewFrame = CTkFrame(self.topviewframe)
        self.admin.initUI(self.adminViewFrame)
        self.nbk.add(self.adminViewFrame, "Admin", order=Module.LOW_PRIORITY, image=self.admin_tab_img)

    def openScriptModule(self):
        """Open the script window"""
        self.scriptManager = ScriptManager()
        self.scriptManager.initUI(self)
        self.wait_window(self.scriptManager.app)

    def initUI(self):
        """
        initialize all the main windows objects. (Bar Menu, contextual menu, treeview, editing pane)
        """
        if self.nbk is not None:
            self.refreshUI()
            return
        self.nbk = ButtonNotebook(self, self.tabSwitch, self.beforeTabSwitch)
        self.panedTerminals = ttk.PanedWindow(self.nbk, orient="vertical")
        self.topviewframe = CTkFrame(self.panedTerminals)
        self.terminals = TerminalsWidget(self.panedTerminals, self,  height=200)
        
        self.initMainView()
        self.initAdminView()
        self.initCommandsView()
        self.initScanView()
        self.initSettingsView()

        for module in self.modules:
            module["view"] = CTkFrame(self.topviewframe)
        for module in self.modules:
            self.nbk.add(module["view"], module["name"].strip(), order=module["object"].__class__.order_priority, image=module["img"])
        self.terminals.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=1)
        self.topviewframe.pack(side=tk.TOP, fill=tk.BOTH, expand=1)

        self.panedTerminals.add(self.topviewframe, weight=1)
        self.panedTerminals.add(self.terminals, weight=0)
        
        self.panedTerminals.pack(fill=tk.BOTH, expand=True)
        

        
        self.nbk.pack(fill=tk.BOTH, expand=1)
        self.terminals.open_terminal()


    def refreshUI(self):
        for widget in self.viewframe.winfo_children():
            widget.destroy()
        
        self.left_pane.update()
        self.after(50, lambda: self.paned.paneconfigure(self.left_pane, height=self.left_pane.winfo_reqheight()))
        
        self.treevw.refresh()
        self.treevw.filter_empty_nodes()
        self.statusbar.refreshTags(Settings.getTags(ignoreCache=True))
        #self.terminals.open_terminal()
        # self.nbk.select("Main View")

    def open_terminal(self, iid, title):
        self.terminals.open_terminal(iid, title)

    def execute_in_terminal(self, title, commandline):
        iid = str(uuid.uuid4())
        self.terminals.open_terminal(iid, title, enable_trap=False)
        self.terminals.launch_in_terminal(iid, commandline, use_pollex=False)

    def launch_in_terminal(self, iid, title, commandline, use_pollex=True):
        if iid is None:
            iid = str(uuid.uuid4())
        self.terminals.open_terminal(iid, title, enable_trap=use_pollex)
        self.terminals.launch_in_terminal(iid, commandline, use_pollex=use_pollex)
        return iid

    def launch_tool_in_terminal(self, tool_model, command):
        self.terminals.open_terminal(str(tool_model.check_iid)+"|"+str(tool_model.getId()), ToolController(tool_model).getDetailedString())
        self.terminals.launch_in_terminal(str(tool_model.check_iid)+"|"+str(tool_model.getId()), command)

    def open_ro_terminal(self, check_iid, title, tool_controller, scanManager):
        self.terminals.open_ro_terminal(check_iid, title, tool_controller, scanManager)

    def open_any_terminal(self, iid, title, tool_controller, scanManager):
        self.terminals.open_any_terminal(iid, title, tool_controller, scanManager)
        
    def textSearchChanged(self, event=None):
        """Called when the textual search bar is modified. Change settings
        Args:
            event: not used but mandatory
        """
        self.settings.local_settings["textsearch"] = int(self.textSearchVal.get())
        self.settings.saveLocalSettings()

    def keepParentsChanged(self, event=None):
        """Called when the keep parent switch is modified. Change settings
        Args:
            event: not used but mandatory
        """
        self.settings.local_settings["keep_parents"] = int(self.keep_parents_val.get())
        self.settings.saveLocalSettings()
    
    def newSearch(self, _event=None, histo=True, text_search_allowed=True):
        """Called when the searchbar is validated (click on search button or enter key pressed).
        Perform a filter on the main treeview.
        Args:
            _event: not used but mandatory"""
        self.searchMode = True
        filterStr = self.searchBar.get()
        if filterStr.strip() == "":
            self.endSearch()
            return
        self.settings.reloadSettings()
        success = self.treevw.filterTreeview(filterStr, self.settings, text_search_allowed)
        self.searchMode = (success and filterStr.strip() != "")
        if success:
            if histo:
                histo_filters = self.settings.local_settings.get("histo_filters", [])
                if filterStr.strip() != "":
                    histo_filters.insert(0, filterStr)
                    if len(histo_filters) > 10:
                        histo_filters = histo_filters[:10]
                    self.settings.local_settings["histo_filters"] = histo_filters
                    self.settings.saveLocalSettings()
            if self.helpFrame is not None:
                self.helpFrame.destroy()
                self.helpFrame = None

    def statusbarClicked(self, name):
        """Called when a button in the statusbar tag is clicked.
        filter the treeview to match the status bar tag clicked and enforce select of main view
        Args:
            name: not used but mandatory"""
        # get the index of the mouse click
        datamanager = DataManager.getInstance()
        taggeds = datamanager.get("tags", "*")
        tagged_items = []
        tagged_types = set()
        for tagged in taggeds:
            for tag in tagged.tags:
                tag = TagInfos(tag)
                if tag.name == name:
                    tagged_items.append(tagged)
                    tagged_types.add(tagged.item_type)
        for module in self.modules:
            for tagged_type in tagged_types:
                if tagged_type.lower() in getattr(module["object"], "classes", []):
                    if hasattr(module["object"], "statusbarClicked"):
                        self.nbk.select(module["name"].strip())
                        module["object"].statusbarClicked(name)
                    return
        # default 
        self.nbk.select("Main View")
        self.searchTaggedBy(name)

    def modelToView(self, collection, model):
        """Return the view of a model"""
        if collection.lower()[-1] != "s":
            collection = collection.lower() + "s"
        for module in self.modules:
            if collection.lower() == getattr(module["object"], "coll_name", "").lower() or collection.lower() in getattr(module["object"], "classes", ""):
                return module["object"].modelToView(collection, model)
        return self.treevw.modelToView(collection, model)
    
    def search(self, filter_str):
        self.nbk.select("Main View")
        
        self.searchBar.delete(0, tk.END)
        self.searchBar.insert(tk.END, filter_str)
        self.newSearch(histo=False, text_search_allowed=False)

    def searchTaggedBy(self, tag_name):
        self.nbk.select("Main View")
        apiclient = APIClient.getInstance()
        searcher = apiclient.searchTaggedBy(tag_name)
        if searcher.get("success", True):
            self.searchMode = True
            self.treevw.doFilterTreeview(searcher, True, keep_parents=self.settings.local_settings.get("keep_parents", True))
            
    def endSearch(self):
        """
        Called when the reset button of the status bar is clicked.
        """
        self.searchMode = False
        self.searchBar.reset()
        self.treevw.unfilterAll()

    def refreshView(self, _event=None):
        """
        Reload the currently opened tab
        Args:
            _event: not used but mandatory
        """
        setViewOn = None
        nbkOpenedTab = self.nbk.getOpenTabName()
        activeTw = None
        if nbkOpenedTab == "Main View":
            activeTw = self.treevw
        elif nbkOpenedTab == "Commands":
            activeTw = self.commandsTreevw
        elif nbkOpenedTab == "Scan":
            self.scanManager.initUI(self.scanViewFrame)
        elif nbkOpenedTab == "Settings":
            self.settings.reloadUI()
        else:
            for module in self.modules:
                if nbkOpenedTab.strip().lower() == module["name"].strip().lower():
                    module["object"].open(module["view"], self.topviewframe, self.treevw)
        if activeTw is not None:
            if len(activeTw.selection()) == 1:
                setViewOn = activeTw.selection()[0]
            activeTw.refresh(force=True)
            self.statusbar.refreshTags(Settings.getTags(ignoreCache=True))

            activeTw.filter_empty_nodes()
        if setViewOn is not None:
            try:
                activeTw.see(setViewOn)
                activeTw.focus(setViewOn)
                activeTw.selection_set(setViewOn)
                activeTw.openModifyWindowOf(setViewOn)
            except tk.TclError:
                pass

    def resetUnfinishedTools(self):
        """
        Reset all running tools to a ready state.
        """
        apiclient = APIClient.getInstance()
        if apiclient.getCurrentPentest() != "":
            utils.resetUnfinishedTools()
            self.treevw.load()

    def wrapperTestLocalTools(self):
        results = self.testLocalTools()
        dialog = ChildDialogToolsInstalled(results)
        if dialog.rvalue is not None:
            self.settings.local_settings["my_commands"] = dialog.rvalue
            self.settings.saveLocalSettings()

    def testLocalTools(self):
        """ test local binary path with which"""
        apiclient = APIClient.getInstance()
        self.settings.reloadLocalSettings()
        plugins = apiclient.getPlugins()
        results = {"successes":[], "failures":[]}
        expanded_bin_paths = []
        for plugin in plugins:
            if plugin["plugin"] == "Default":
                continue
            my_commands = self.settings.local_settings.get("my_commands", {})
            my_commands = {} if my_commands is None else my_commands
            bin_path = my_commands.get(plugin["plugin"])
            expanded_bin_paths.append(bin_path)
            expanded_bin_paths += plugin["default_bin_names"]
        expanded_paths = utils.which_expand_aliases(expanded_bin_paths)
        for plugin in plugins:
            if plugin["plugin"] == "Default":
                continue
            my_commands = self.settings.local_settings.get("my_commands", {})
            my_commands = {} if my_commands is None else my_commands
            bin_path = my_commands.get(plugin["plugin"])
            expanded_bin_path = expanded_paths.get(bin_path)
            if bin_path is None or bin_path == "" or expanded_bin_path is None:
                default_bin_names = plugin["default_bin_names"]
                found_matching = False
                for default_bin_name in default_bin_names:
                    if expanded_paths.get(default_bin_name):
                        plugin["bin_path"] = default_bin_name
                        bin_path = default_bin_name
                        my_commands[plugin["plugin"]] = default_bin_name
                        results["successes"].append({"title":"Success", "plugin":plugin, "bin_path":bin_path,  "default_bin":plugin["default_bin_names"], "msg":f"The local settings for {plugin['plugin']} is valid. ({bin_path})."})
                        found_matching = True
                        break
                if found_matching == False:
                    results["failures"].append({"title":"Invalid binary path", "plugin":plugin, "bin_path":bin_path,  "default_bin":plugin["default_bin_names"], "msg":f"The local settings for {plugin['plugin']} is not recognized. ({bin_path})."})
            elif expanded_bin_path is not None:
                results["successes"].append({"title":"Success", "plugin":plugin, "bin_path":bin_path,  "default_bin":plugin["default_bin_names"], "msg":f"The local settings for {plugin['plugin']} is valid. ({bin_path})."})
            else:
                results["failures"].append({"title":"Invalid binary path", "plugin":plugin, "bin_path":bin_path,  "default_bin":plugin["default_bin_names"], "msg":f"The local settings for {plugin['plugin']} is not recognized. ({bin_path})."})
        self.settings.local_settings["my_commands"] = my_commands
        self.settings.saveLocalSettings()
        return results
    

    def exportCommands(self):
        """
        Dump pollenisator from database to an archive file gunzip.
        """
        dialog = ChildDialogQuestion(self, "Ask question", "Do you want to export your commands or Worker's commands.", ["My commands", "Worker"])
        self.wait_window(dialog.app)
        apiclient = APIClient.getInstance()
        res, msg = apiclient.exportCommands(self)
        if res:
            tkinter.messagebox.showinfo(
                "Export pollenisator database", "Export completed in "+str(msg))
        else:
            tkinter.messagebox.showinfo(msg)
    
    def exportCheatsheet(self):
        """
        Dump pollenisator from database to an archive file gunzip.
        """
        apiclient = APIClient.getInstance()
        res, msg = apiclient.exportCheatsheet(self)
        if res:
            tkinter.messagebox.showinfo(
                "Export cheatsheet database", "Export completed in "+str(msg))
        else:
            tkinter.messagebox.showinfo(msg)

   

    def findUnscannedPorts(self):
        ports = Port.fetchObjects({})
        datamanager = DataManager.getInstance()
        for port in ports:
            port_key = port.getDbKey()
            res = datamanager.find("tools", port_key, multi=False)
            if res is None:
                port.setTags(["unscanned"])

    def importCommands(self, name=None):
        """
        Import a pollenisator archive file gunzip to database.
        Args:
            name: The filename of the gunzip command table exported previously
        Returns:
            None if name is None and filedialog is closed
            True if commands successfully are imported
            False otherwise.
        """
        filename = ""
        if name is None:
            f = tkinter.filedialog.askopenfilename(parent=self, defaultextension=".json")
            if f is None:  # asksaveasfile return `None` if dialog closed with "cancel".
                return
            filename = str(f)
        else:
            filename = name
        try:
            dialog = ChildDialogQuestion(self, "Ask question", "Do you want to import these commands for you or the worker.", ["Me", "Worker"])
            self.wait_window(dialog.app)
            apiclient = APIClient.getInstance()
            success = apiclient.importCommands(filename, forWorker=dialog.rvalue == "Worker")
            self.commandsTreevw.refresh()
        except IOError:
            tkinter.messagebox.showerror(
                "Import commands", "Import failed. "+str(filename)+" was not found or is not a file.")
            return False
        if not success:
            tkinter.messagebox.showerror("Command import", "Command import failed")
        else:
            tkinter.messagebox.showinfo("Command import", "Command import completed")
        return success

    def importCheatsheet(self, name=None):
        """
        Import a pollenisator cheatsheet file json to database.
        Args:
            name: The filename of the json command table exported previously
        Returns:
            None if name is None and filedialog is closed
            True if commands successfully are imported
            False otherwise.
        """
        filename = ""
        if name is None:
            f = tkinter.filedialog.askopenfilename(parent=self, defaultextension=".json")
            if f is None:  # asksaveasfile return `None` if dialog closed with "cancel".
                return
            filename = str(f)
        else:
            filename = name
        try:
            apiclient = APIClient.getInstance()
            success = apiclient.importCheatsheet(filename)
        except IOError:
            tkinter.messagebox.showerror(
                "Import Cheatsheet", "Import failed. "+str(filename)+" was not found or is not a file.")
            return False
        if not success:
            tkinter.messagebox.showerror("Cheatsheet import", "Cheatsheet import failed")
        else:
            tkinter.messagebox.showinfo("Cheatsheet import", "Cheatsheet import completed")
        return success

    def importDefectTemplates(self, name=None):
        """
        Import defect templates from a json
        Args:
            name: The filename of the json containing defect templates
        Returns:
            None if name is None and filedialog is closed
            True if defects successfully are imported
            False otherwise.
        """
        filename = ""
        if name is None:
            f = tkinter.filedialog.askopenfilename(defaultextension=".json")
            if f is None:  # asksaveasfile return `None` if dialog closed with "cancel".
                return
            filename = str(f)
        else:
            filename = name
        try:
            apiclient = APIClient.getInstance()
            success = apiclient.importDefectTemplates(filename)
        except IOError:
            tkinter.messagebox.showerror(
                "Import defects templates", "Import failed. "+str(filename)+" was not found or is not a file.")
            return False
        if not success:
            tkinter.messagebox.showerror("Defects templates import", "Defects templatest failed")
        else:
            tkinter.messagebox.showinfo("Defects templates import", "Defects templates completed")
        return success

    def openPentestsWindow(self, _event=None, pentests=None):
        """
        Open Pentest dialog window
        Args:
            _event: Not used but mandatory
        Returns:
            None if no database were selected
            datababase name otherwise
        """
        dialog = ChildDialogPentests(self, pentests)
        try:
            dialog.wait_window(dialog)
        except tk.TclError:
            pass
        if dialog.rvalue is not None:
            self.openPentest(dialog.rvalue)
        return dialog.rvalue


    def newPentest(self, pentestName, pentest_type, start_date, end_date, scope, settings, pentesters):
        """
        Register the given pentest name into database and opens it.

        Args:
            pentestName: The pentest database name to register in database.
        """
        succeed = False
        if pentestName is not None:
            apiclient = APIClient.getInstance()
            succeed, msg = apiclient.registerPentest(pentestName, pentest_type, start_date, end_date, scope, settings, pentesters)
            if not succeed:
                tkinter.messagebox.showinfo("Forbidden", msg)
        return succeed

    def closePentest(self, close_terminals=True):
        """
        Close the current pentest and refresh the treeview.
        """
        apiclient = APIClient.getInstance()
        apiclient.dettach(self)
        if self.terminals is not None and close_terminals:
            self.terminals.onClosing()
        
        if self.scanManager is not None:
            self.scanManager.onClosing()
        
            
        for module in self.modules:
            if callable(getattr(module["object"], "onClosing", None)):
                module["object"].onClosing()

    def openPentest(self, filename=""):
        """
        Open the given database name. Loads it in treeview.

        Args:
            filename: the pentest database name to load in application. If "" is given (default), will refresh the already opened database if there is one.
        """
        pentestName = None
        apiclient = APIClient.getInstance()

        if filename == "" and apiclient.getCurrentPentest() != "":
            pentestName = apiclient.getCurrentPentest()
        elif filename != "":
            pentestName = filename.split(".")[0].split("/")[-1]
        if pentestName is not None:
            self.closePentest(close_terminals=False)
            first_use_detected = self.detectFirstUse()
            res = apiclient.setCurrentPentest(pentestName, first_use_detected)
            if not res:
                tk.messagebox.showerror("Connection failed", "Could not connect to "+str(pentestName))
                return
            DataManager.getInstance().openPentest(pentestName)
            self.initUI()
            self.settings.reloadSettings()
            self.statusbar.refreshTags(Settings.getTags(ignoreCache=True))
            self.sio.emit("registerForNotifications", {"token":apiclient.getToken(), "pentest":pentestName})
            self.refresh_tabs()
            self.nbk.select("Dashboard")

    def refresh_tabs(self):
        apiclient = APIClient.getInstance()
        if apiclient.isAdmin():
            self.nbk.add(self.adminViewFrame, "Admin", order=Module.LAST_PRIORITY, image=self.admin_tab_img)
        else:
            self.nbk.delete("Admin")
        pentest_type = self.settings.getPentestType()
        for module in self.modules:
            pentest_type_allowed = pentest_type.lower() in module["object"].__class__.pentest_types
            all_are_authorized = "all" in module["object"].__class__.pentest_types
            module_need_admin = module["object"].__class__.need_admin
            is_admin = apiclient.isAdmin()
            if (pentest_type_allowed or all_are_authorized) and (is_admin or not module_need_admin):
                self.nbk.add(module["view"], module["name"].strip(), order=module["object"].__class__.order_priority, image=module["img"])
            else:    
                self.nbk.delete(module["name"])

   
    def importExistingTools(self, _event=None):
        """
        Ask user to import existing files to import.
        """
        dialog = ChildDialogFileParser(self)
        self.wait_window(dialog.app)

    def detectFirstUse(self):
        detector = os.path.join(utils.getConfigFolder(),".first_use")
        if os.path.exists(detector):
            return False
        with open(detector, mode="w") as f:
            f.write("")
        return True



"""
Simple calendar using ttk Treeview together with calendar and datetime
classes.
Adapted from github.
"""
import calendar

import tkinter as tk
import tkinter.font as tkFont
import tkinter.ttk as ttk
from customtkinter import *
import sys
from datetime import datetime
from datetime import timedelta

def get_calendar(locale, fwday):
    # instantiate proper calendar class
    if locale is None:
        return calendar.TextCalendar(fwday)
    else:
        return calendar.LocaleTextCalendar(fwday, locale)

class ChildDialogDate:
    def __init__(self, parent=None):
        self.parent = parent
        self.app = CTkToplevel(self.parent)
        self.app.attributes("-type", "dialog")
        self.app.title("Choose a date")
        self.app.resizable(False, False)
        self.app.bind("<Escape>", self.destroy)
        self.appFrame = CTkFrame(self.app)
        self.initCalendar(self.app)
        self.rvalue = None
        self.parent = parent
        self.appFrame.pack(expand=1, fill='both')
        try:
            self.app.wait_visibility()
            self.app.transient(parent)
            self.app.focus_force()
            self.app.grab_set()
            self.app.lift()
        except tk.TclError:
            pass


    def show(self):
        """Start displaying this window."""
        self.app.update()

    def destroy(self, _event=None):
        """
        Close the window.
        """
        # send the data to the parent
        self.rvalue = None
        self.app.destroy()

    def initCalendar(self, master=None, **kw):
        """
        WIDGET-SPECIFIC OPTIONS

            locale, firstweekday, year, month, selectbackground,
            selectforeground
        """
        # remove custom options from kw before initializating CTkFrame
        fwday = kw.pop('firstweekday', calendar.MONDAY)
        year = kw.pop('year', datetime.now().year)
        month = kw.pop('month', datetime.now().month)
        locale = kw.pop('locale', None)
        sel_bg = kw.pop('selectbackground', '#ecffc4')
        sel_fg = kw.pop('selectforeground', '#05640e')
        self.master = master
        self._date = datetime(year, month, 1)
        self._selection = None # no date selected


        self._cal = get_calendar(locale, fwday)

        self.__setup_styles()       # creates custom styles
        self.__place_widgets()      # pack/grid used widgets
        self.__config_calendar()    # adjust calendar columns and setup tags
        # configure a canvas, and proper bindings, for selecting dates
        self.__setup_selection(sel_bg, sel_fg)

        # store items ids, used for insertion later
        self._items = [self._calendar.insert('', 'end', values='')
                            for _ in range(6)]
        # insert dates in the currently empty calendar
        self._build_calendar()

        # set the minimal size for the widget
        self._calendar.bind('<Map>', self.__minsize)

    def __setitem__(self, item, value):
        if item in ('year', 'month'):
            raise AttributeError("attribute '%s' is not writeable" % item)
        elif item == 'selectbackground':
            self._canvas['background'] = value
        elif item == 'selectforeground':
            self._canvas.itemconfigure(self._canvas.text, item=value)
        else:
            CTkFrame.__setitem__(self, item, value)

    def __getitem__(self, item):
        if item in ('year', 'month'):
            return getattr(self._date, item)
        elif item == 'selectbackground':
            return self._canvas['background']
        elif item == 'selectforeground':
            return self._canvas.itemcget(self._canvas.text, 'fill')
        else:
            r = ttk.tclobjs_to_py({item: CTkFrame.__getitem__(self, item)})
            return r[item]

    def __setup_styles(self):
        # custom ttk styles
        style = ttk.Style(self.master)
        arrow_layout = lambda dir: (
            [('Button.focus', {'children': [('Button.%sarrow' % dir, None)]})]
        )
        style.layout('L.TButton', arrow_layout('left'))
        style.layout('R.TButton', arrow_layout('right'))

    def __place_widgets(self):
        # header frame and its widgets
        hframe = CTkFrame(self.appFrame)
        lbtn = ttk.Button(hframe, style='L.TButton', command=self._prev_month)
        rbtn = ttk.Button(hframe, style='R.TButton', command=self._next_month)
        self._header = CTkLabel(hframe, width=15, anchor='center')
        # the calendar
        self._calendar = ttk.Treeview(self.app, show='', selectmode='none', height=7)

        # pack the widgets
        hframe.pack(in_=self.appFrame, side='top', pady=4, anchor='center')
        lbtn.grid(in_=hframe)
        self._header.grid(in_=hframe, column=1, row=0, padx=12)
        rbtn.grid(in_=hframe, column=2, row=0)
        self._calendar.pack(expand=1, fill='both', side='bottom')

    def __config_calendar(self):
        cols = self._cal.formatweekheader(3).split()
        self._calendar['columns'] = cols
        self._calendar.tag_configure('header', background='grey90')
        self._calendar.insert('', 'end', values=cols, tag='header')
        # adjust its columns width
        font = tkFont.Font()
        maxwidth = max(font.measure(col) for col in cols)
        for col in cols:
            self._calendar.column(col, width=maxwidth, minwidth=maxwidth,
                anchor='e')

    def __setup_selection(self, sel_bg, sel_fg):
        self._font = tkFont.Font()
        self._canvas = canvas = tk.Canvas(self._calendar,
            background=sel_bg, borderwidth=0, highlightthickness=0)
        canvas.text = canvas.create_text(0, 0, fill=sel_fg, anchor='w')

        canvas.bind('<ButtonPress-1>', lambda evt: canvas.place_forget())
        self._calendar.focus_set()
        self._calendar.bind('<Configure>', lambda evt: canvas.place_forget())
        self._calendar.bind('<ButtonPress-1>', self._pressed)
    def __minsize(self, evt):
        width, height = self._calendar.master.geometry().split('x')
        height = height[:height.index('+')]
        self._calendar.master.minsize(int(width), int(height))

    def _build_calendar(self):
        year, month = self._date.year, self._date.month

        # update header text (Month, YEAR)
        header = self._cal.formatmonthname(year, month, 0)
        self._header.configure(text=header.title())

        # update calendar shown dates
        cal = self._cal.monthdayscalendar(year, month)
        for indx, item in enumerate(self._items):
            week = cal[indx] if indx < len(cal) else []
            fmt_week = [('%02d' % day) if day else '' for day in week]
            self._calendar.item(item, values=fmt_week)

    def _show_selection(self, text, bbox):
        """Configure canvas for a new selection."""
        x, y, width, height = bbox

        textw = self._font.measure(text)

        canvas = self._canvas
        canvas.configure(width=width, height=height)
        canvas.coords(canvas.text, width - textw, height / 2 - 1)
        canvas.itemconfigure(canvas.text, text=text)
        canvas.place(in_=self._calendar, x=x, y=y)

    # Callbacks

    def _pressed(self, evt):
        """Clicked somewhere in the calendar."""
        x, y, widget = evt.x, evt.y, evt.widget
        item = widget.identify_row(y)
        column = widget.identify_column(x)

        if not column or not item in self._items:
            # clicked in the weekdays row or just outside the columns
            return

        item_values = widget.item(item)['values']
        if not len(item_values): # row is empty for this month
            return

        text = item_values[int(column[1]) - 1]
        if not text: # date is empty
            return

        bbox = widget.bbox(item, column)
        if not bbox: # calendar not visible yet
            return

        # update and then show selection
        text = '%02d' % text
        if self._selection is not None:
            if self._selection[0] == text and self._selection[1] == item and  self._selection[2] == column:
                self.rvalue = self.selection
                self.app.destroy()
                return
        self._selection = (text, item, column)
        self._show_selection(text, bbox)

    def _prev_month(self):
        """Updated calendar to show the previous month."""
        self._canvas.place_forget()

        self._date = self._date - timedelta(days=1)
        self._date = datetime(self._date.year, self._date.month, 1)
        self._build_calendar() # reconstuct calendar

    def _next_month(self):
        """Update calendar to show the next month."""
        self._canvas.place_forget()

        year, month = self._date.year, self._date.month
        self._date = self._date + timedelta(
            days=calendar.monthrange(year, month)[1] + 1)
        self._date = datetime(self._date.year, self._date.month, 1)
        self._build_calendar() # reconstruct calendar

    # Properties

    @property
    def selection(self):
        """Return a datetime representing the current selected date."""
        if not self._selection:
            return None

        year, month = self._date.year, self._date.month
        return datetime(year, month, int(self._selection[0]))




"""Widget with a button that display an helping message when hovered"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from pollenisatorgui.core.forms.form import Form
from PIL import ImageTk, Image
from pollenisatorgui.core.components.utils import getIcon


class FormHelper(Form):
    """
    Form field representing a helper button.
    Default setted values: 
        state="readonly"
        if pack : padx = pady = 5, side = "right"
        if grid: row = column = 0 sticky = "west"
        entry "width"=  20
    """
    img_class = None

    def __init__(self, name, text, **kwargs):
        """
        Constructor for a form button

        Args:
            name: the helper name. Should not matter as it does not return data
            text: the helping message to be displayed
            kwargs: same keyword args as you would give to CTkLabel
        """
        super().__init__(name)
        FormHelper.img_class = CTkImage(
            Image.open(getIcon("help.png")))
        self.text = text
        self.kwargs = kwargs
        self.lbl = None
        self.tw = None

    def constructView(self, parent):
        """
        Create the button view inside the parent view given

        Args:
            parent: parent form panel.
        """
        self.lbl = CTkLabel(
            parent.panel, text="", image=FormHelper.img_class)
        self.lbl.bind("<Enter>", self.enter)
        self.lbl.bind("<Leave>", self.close)
        if parent.gridLayout:
            self.lbl.grid(row=self.getKw("row", 0), column=self.getKw(
                "column", 0), sticky=self.getKw("sticky", tk.W))
        else:
            self.lbl.pack(side=self.getKw("side", "right"), padx=self.getKw(
                "padx", 5), pady=self.getKw("pady", 5))

    def enter(self, _event=None):
        """Callback for the <Enter> event
        Starts displaying the help message
        Args:
            _event: not used but mandatory
        """
        x = y = 0
        x, y, _, _ = self.lbl.bbox("insert")
        x += self.lbl.winfo_rootx() + 25
        y += self.lbl.winfo_rooty() + 20
        # creates a toplevel window
        self.tw = CTkToplevel(self.lbl)
        self.tw.attributes("-type", "dialog")
        self.tw.title("Helper")
        # Leaves only the label and removes the app window
        self.tw.wm_overrideredirect(True)
        self.tw.wm_geometry("+%d+%d" % (x, y))
        label = CTkLabel(self.tw, text=self.text, justify='left',
                         fg_color=('light yellow', 'saddle brown'))
        label.pack(ipadx=1)

    def close(self, _event=None):
        """Callback for the <Leave> event
        Stops displaying the help message
        Args:
            _event: not used but mandatory
        """
        if self.tw:
            self.tw.destroy()



"""View for scope list object. Present an multi insertion form to user when interacted with."""

from pollenisatorgui.core.views.viewelement import ViewElement


class MultipleScopeView(ViewElement):
    """View for scope list object. Present an multi insertion form to user when interacted with."""


    def openInsertWindow(self, check_scope=True):
        """
        Creates a tkinter form using Forms classes. This form aims to insert many new Scopes
        """
        modelData = self.controller.getData()
        top_panel = self.form.addFormPanel(grid=True)
        top_panel.addFormLabel("Insert Scopes")
        top_panel.addFormHelper(
            "Add Network IP 'XXX.XXX.XXX.XXX/XX' or domain names to the pentest's scope\nEach network/domain name must on a separated line", column=1)
        top_panel = self.form.addFormPanel()
        top_panel.addFormText("Scopes", r"^(?!https?://).+$", modelData.get("scope", ""), None, side="top")
        top_panel = self.form.addFormPanel()
        top_panel.addFormSwitch(
            "Split", "Make ip range split into /24 at most", False, side="left")
        top_panel.addFormHelper(
            "This will split big ranges in smaller one. May slow pollenisator opening.\nBut you will have scan results quicker", side="left")
        top_panel = self.form.addFormPanel()
        top_panel.addFormSwitch(
            "Settings", "Check domain scope", check_scope, side="left")
        top_panel.addFormHelper(
            "If you want to insert a brand new scope in the scope, you will have to disable this security", side="left")
        self.form.addFormHidden("wave", modelData["wave"])
        self.completeInsertWindow()

    # def addChildrenBaseNodes(self, newNode):
    #     """
    #     Add to the given node from a treeview the mandatory childrens.
    #     For a scope it is the tools parent node and the ips parent node.

    #     Args:
    #         newNode: the newly created node we want to add children to.
    #     """
    #     ScopeView(self.appliTw, self.appliViewFrame, self.mainApp,
    #               self.controller).addChildrenBaseNodes(newNode)
        
    def getAdditionalContextualCommands(self):
        return {"Insert Scopes":self.openInsertWindow}


"""View for user object. Handle node in treeview and present forms to user when interacted with."""

from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.views.viewelement import ViewElement
from tkinter import TclError
import tkinter as tk

class UserView(ViewElement):
    """View for user object. Handle node in treeview and present forms to user when interacted with.
    Attributes:
        icon: icon name to show in treeview. Icon filename must be in icon directory.
    """

    icon = 'user.png'
    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to update or delete an existing User
        """
        self.form.clear()
        user_data = self.controller.getData()
        panel = self.form
        panel_info = panel.addFormPanel(grid=True)
        panel_info.addFormLabel("Domain")
        panel_info.addFormStr("Domain", "", user_data.get("domain", ""), status="readonly", row=0, column=1)
        panel_info.addFormLabel("Username", row=1)
        panel_info.addFormStr("Username", "", user_data.get("username", ""), status="readonly", row=1, column=1)
        panel_info.addFormLabel("Password", row=2)
        panel_info.addFormStr("Password", "", user_data.get("password", ""), status="readonly", row=2, column=1)
        panel.addFormLabel("Desc", text=f"Desc : {user_data.get('desc', '')}" , side="top")
        if user_data.get("infos", {}).get("secrets", []):
            panel.addFormLabel("Secrets", side="top")
            panel.addFormTreevw("Secrets", ("Secret", ""), [(s, "") for s in user_data.get("infos", {}).get("secrets", [])], side="top")
        groups = user_data.get('groups', []) 
        if groups is None:
            groups = []
        panel.addFormTreevw("Groups", ("Group",), [ [x] for x in groups], side="top")
        self.completeModifyWindow(editable=False)
        
    def openInsertWindow(self):
        """
        Creates a tkinter form using Forms classes. This form aims to insert a new User
        """
        # modelData = self.controller.getData()
        # top_panel = self.form.addFormPanel(grid=True)
        # top_panel.addFormLabel("Start date")
        # top_panel.addFormDate("Start date", self.mainApp, "01/01/2000 00:00:00", column=1)
        # top_panel.addFormLabel("End date", row=1)
        # top_panel.addFormDate(
        #     "End date", self.mainApp, "31/12/2099 00:00:00", row=1, column=1)
        # top_panel.addFormHelper(
        #     "Auto scan will only launch if current time fits this user", row=1, column=2)
        # # added in getEmptyModel
        # self.form.addFormHidden("waveName", modelData["wave"])

        # self.completeInsertWindow()
        raise NotImplementedError("UserView.openInsertWindow not implemented")

    def getAdditionalContextualCommands(self):
        return {}

    def addInTreeview(self, parentNode=None, **kwargs):
        """Add this view in treeview. Also stores infos in application treeview.
        Args:
            parentNode: if None, will calculate the parent. If setted, forces the node to be inserted inside given parentNode.
            _addChildren: not used here
        """

        self.appliTw.views[str(self.controller.getDbId())] = {"view": self}
        tags = self.controller.getTags()
        data = self.controller.getData()
        if parentNode is None:
            parentNode = data.get("domain")+" Users"
        try:
            self.appliTw.insert(
                "", 0, parentNode, text=data.get("domain")+" Users", image=self.getClassIcon())
        except TclError as e:  #  trigger if tools list node already exist
            pass
        try:
            self.appliTw.insert(parentNode, "end", str(
                self.controller.getDbId()), text=str(self.controller.getModelRepr()), tags=tags, image=self.getClassIcon())
        except TclError as e:
            pass
        if "hidden" in tags:
            self.hide("tags")
        if self.mainApp.settings.is_checklist_view():
            self.hide("checklist_view")

   


"""Controller for command object. Mostly handles conversion between mongo data and python objects"""
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.controllers.controllerelement import ControllerElement
from pollenisatorgui.core.components.settings import Settings
import bson

class CommandController(ControllerElement):
    """Inherits ControllerElement
    Controller for command object. Mostly handles conversion between mongo data and python objects"""
    def doUpdate(self, values):
        """
        Update the command represented by this self.model in database with the given values.

        Args:
            values: A dictionary crafted by CommandView containg all form fields values needed.

        Returns:
            The mongo ObjectId _id of the updated command document.
        """
        # Get form values
        self.model.bin_path = values.get(
            "Bin path", self.model.bin_path)
        local_bin_path = values.get(
            "My binary path", None)
        if local_bin_path is not None:
            settings = Settings()
            settings.local_settings["my_commands"] = settings.local_settings.get("my_commands", {})
            settings.local_settings["my_commands"][self.model.name] = local_bin_path
            settings.saveLocalSettings()
        self.model.plugin = values.get(
            "Plugin", self.model.plugin)
        self.model.text = values.get("Command line options", self.model.text)
        self.model.timeout = str(values.get("Timeout", self.model.timeout))
        self.model.owners = values.get("owners", [])
        # Update in database
        self.model.update()

    def doInsert(self, values):
        """
        Insert the command represented by this model in the database with the given values.

        Args:
            values: A dictionary crafted by CommandView containg all form fields values needed.

        Returns:
            {
                'Command': The Command object associated
                'nbErrors': The number of objects that has not been inserted in database due to errors.
            }
        """
        # Get form values
        bin_path = values["Bin path"]
        plugin = values["Plugin"]
        text = values["Command line options"]
        name = values["Name"]
        indb = values["indb"]
        timeout = values["Timeout"]
        owners = values["owners"]
        #types = [k for k, v in types.items() if v == 1]
        self.model.initialize(name, bin_path, plugin, 
                              text,  indb, None, timeout, owners)
        # Insert in database
        ret, iid = self.model.addInDb()
        if not ret:
            # command failed to be inserted, a duplicate exists
            # return None as inserted_id and 1 error
            return None, 1
        # Fetch the instance of this self.model now that it is inserted.
        return ret, (0, iid)  # 0 errors

    
    def getType(self):
        """Return a string describing the type of object
        Returns:
            "command" """
        return "command"

    def actualize(self):
        """Ask the model to reload its data from database
        """
        if self.model is not None:
            self.model = self.model.__class__.fetchObject(
                {"_id": bson.ObjectId(self.model.getId())}, self.model.indb)

    def addToMyCommands(self, event=None):
        """Add command to current user's commands
        """
        self.model.addToMyCommands()

    def removeFromMyCommands(self, event=None):
        """Remove command from current user's commands
        """
        self.model.removeFromMyCommands()

    def isMine(self):
        return self.model.isMine()




"""Controller for command object. Mostly handles conversion between mongo data and python objects"""
from turtle import title
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.controllers.controllerelement import ControllerElement
import bson
import json

class CheckItemController(ControllerElement):
    """Inherits ControllerElement
    Controller for CheckItem object. Mostly handles conversion between mongo data and python objects"""
    def doUpdate(self, values, updateInDb=True):
        """
        Update the command represented by this self.model in database with the given values.

        Args:
            values: A dictionary crafted by CheckItemView containg all form fields values needed.

        Returns:
            The mongo ObjectId _id of the updated command document.
        """
        # Get form values
        self.model.title = values.get("Title", self.model.title)
        self.model.description = values.get("Description", self.model.description)
        self.model.check_type = values.get("Check type", self.model.check_type)
        self.model.lvl = values.get("Level", self.model.lvl)
        self.model.ports = values.get("Ports/Services", self.model.ports)
        tag_trigger = values.get("Tag Name", getattr(self.model, "tag_trigger", ""))
        if self.model.lvl.startswith("tag:") and tag_trigger != "":
            self.model.lvl = ":".join(self.model.lvl.split(":")[:2]) +":"+str(tag_trigger)
        self.model.priority = int(values.get("Priority", self.model.priority))
        self.model.max_thread = int(values.get("Shared threads", self.model.max_thread))
        self.model.category = values.get("Category", self.model.category)
        self.model.pentest_types = [x[0] for x in values.get("Pentest types", self.model.pentest_types).items() if x[1]]
        commands_raw = values.get("Commands", [])
        self.model.commands = [str(command[-1]) for command in commands_raw if str(command[-1]) != ""]
        defect_tags = values.get("Defects", [])
        self.model.defect_tags = [(defect[-2], str(defect[-1])) for defect in defect_tags if str(defect[-1]) != ""]
        self.model.script = values.get("Script", "")
        if updateInDb:
            # Update in database
            self.model.update()

    def doInsert(self, values):
        """
        Insert the command represented by this model in the database with the given values.

        Args:
            values: A dictionary crafted by CommandView containg all form fields values needed.

        Returns:
            {
                'Command': The Command object associated
                'nbErrors': The number of objects that has not been inserted in database due to errors.
            }
        """
        # Get form values
        title = values["Title"]
        description = values["Description"]
        ports = values.get("Ports/Services", "")
        lvl = values["Level"]
        priority = values["Priority"]
        max_thread = values.get("Shared threads", 1)
        category = values.get("Category", "")
        commands_raw = values.get("Commands", [])
        defects_raw = values.get("Defects", [])
        check_type = values["Check type"]
        pentest_types = [k for k,v in values["Pentest types"].items() if v]
        step = values["Step"]
        script = values.get("Script", "")
        parent = values.get("Parent", None)
        tag_trigger = values.get("Tag Name", "")
        if lvl.startswith("tag:") and tag_trigger != "":
            lvl = ":".join(lvl.split(":")[:2]) +":"+str(tag_trigger)
        if parent == "":
            parent = None
        commands = []
        for command in commands_raw:
            if len(command) == 3:
                commands.append(str(command[2]))
        defect_tags = []
        for defect in defects_raw:
            if len(defect) == 4:
                defect_tags.append((defect[4],defect[3]))
        self.model.initialize(title, pentest_types=pentest_types, lvl=lvl, ports=ports, priority=priority, max_thread=max_thread, description=description, category=category, check_type=check_type, step=step, parent=parent, commands=commands, defect_tags=defect_tags, script=script,  infos=None)
        # Insert in database
        ret, _ = self.model.addInDb()
        if not ret:
            # command failed to be inserted, a duplicate exists
            # return None as inserted_id and 1 error
            return None, 1
        # Fetch the instance of this self.model now that it is inserted.
        return ret, 0  # 0 errors
    
    def isAuto(self):
        """Return True if this command is an auto command"""
        return self.model.isAuto()

    def getType(self):
        """Return a string describing the type of object
        Returns:
            "checkitem" """
        return "checkitem"

    def getChecks(self):
        """Return check instances that implements this check item
        Returns:
            list of checkInstance objects
        """
        return self.model.getChecks()
    
    def getCategory(self):
        return self.model.category
        
    def getParent(self):
        if self.model is not None:
            return self.model.parent

    def actualize(self):
        """Ask the model to reload its data from database
        """
        if self.model is not None:
            self.model = self.model.__class__.fetchObject({"_id": bson.ObjectId(self.model.getId())})

   


import subprocess
import tkinter
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.application.dialogs.ChildDialogCombo import ChildDialogCombo
from pollenisatorgui.core.application.dialogs.ChildDialogAskText import ChildDialogAskText
import psutil
from pollenisatorgui.core.components.apiclient import APIClient
import os

from pollenisatorgui.scripts.lan.utils import checkPath


def main(apiclient, appli, **kwargs):
   res, path = checkPath(["responder", "Responder.py", "responder.py"])
   if res is None:
      return False, path
   responder_path = path
   APIClient.setInstance(apiclient)
   responder_conf = ""
   if utils.which_expand_alias("locate"):
        resp = subprocess.run("locate Responder.conf", capture_output=True, text=True, shell=True)
        stdout = resp.stdout
        if stdout.strip() == "":
            file = tkinter.filedialog.askopenfilename(title="Locate responder conf file please:", filetypes=[('Config Files', '*.conf')])
            if file:
                responder_conf = file
            else:
                return False, "Responder conf not given"
        else:
            dialog = ChildDialogCombo(None, stdout.split("\n"), displayMsg="Choose your responder config file", width=200)
            dialog.app.wait_window(dialog.app)
            if dialog.rvalue is not None:
                responder_conf = dialog.rvalue.strip()
                cmd = 'sed -i -E "s/(HTTP|SMB) = Off/\\1 = On/gm" '+str(responder_conf)
                if os.geteuid() != 0:
                    cmd = "sudo "+cmd
                appli.launch_in_terminal(None, "sed for responder", cmd, use_pollex=False)
   addrs = psutil.net_if_addrs()
   print(addrs.keys())
   dialog = ChildDialogCombo(None, addrs.keys(), displayMsg="Choose your ethernet device to listen on")
   dialog.app.wait_window(dialog.app)
   if dialog.rvalue is not None:
      cmd = f"{responder_path} -I {dialog.rvalue} -w -d"
      if os.geteuid() != 0:
         cmd = "sudo "+cmd
      appli.launch_in_terminal(kwargs.get("default_target",None), "Responder listening", cmd, use_pollex=False)
   return True, f"Listening responder open"


"""DEPRECATED Hold functions to interact with the search bar. NOW LOCATED IN Filter.py"""

from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.models.ip import Ip
from pollenisatorgui.core.models.port import Port
from pollenisatorgui.core.models.scope import Scope
from pollenisatorgui.core.models.tool import Tool
from pollenisatorgui.core.models.defect import Defect
from pollenisatorgui.core.views.ipview import IpView
from pollenisatorgui.core.views.portview import PortView
from pollenisatorgui.core.views.scopeview import ScopeView
from pollenisatorgui.core.views.toolview import ToolView
from pollenisatorgui.core.views.defectview import DefectView
import re


class Operator:
    def __init__(self, op, arg, exact_match=True):
        self.op = op
        self.arg = arg
        try:
            regex = '^'+self.arg+'$' if exact_match else self.arg
            self.reg = re.compile(regex, re.IGNORECASE | re.MULTILINE)
        except re.error as e:
            raise e

    def getMongo(self):
        key = self.op
        val = {}
        val["$regex"] = self.reg
        return key, val


class ParseError(Exception):
    """To raise parsing error"""


class Search:
    exact_match = True
    classes = {
        "port": {"collection": "ports", "view": PortView, "model": Port, "keywords": ["port", "proto", "service", "product", "ip", "scope", "notes", "tags", "infos."]},
        "ip": {"collection": "ips", "view": IpView, "model": Ip, "keywords": ["ip", "scope", "notes", "tags", "infos."]},
        "scope": {"collection": "scopes", "view": ScopeView, "model": Scope, "keywords": ["wave", "scope", "notes", "tags"]},
        "tool": {"collection": "tools", "view": ToolView, "model": Tool, "keywords": ["name", "dated", "datef", "lvl", "port", "proto", "ip", "scope", "notes", "tags"]},
        "defect": {"collection": "defects", "view": DefectView, "model": Defect, "keywords": ["title", "ease", "impact", "risk", "port", "proto", "ip", "scope", "notes", "tags"]},
    }

    @classmethod
    def help(cls):
        # pylint: disable=anomalous-backslash-in-string
        return """Search examples in match: (python regex)
every ports                       || class:port
every 443 ports                || class:port port:^443$
ports ending with 443      || class:port port:443$
all 80 or 443 ports            || class:port port:^80$|^443$
every port excepted 443  || port:(?!^443$) class:port
ports excepted 80|443     || port:(?!^443$|^80$) class:port
all defects                        || class:defect
every defect with Foo      || class:defect title:foo
all domains                      || class:ip ip:[A-Za-z]
every ip starting as 1.2    || ip:^1\.2 class:ip
every finished tool         || class:tool datef:^(?!None).+
every tool not done         || class:tool datef:None
search an ip infos for ABC info value || class:ip ip.infos.ABC:value
----------------------------------------------------
Search examples in exact match: (python regex)
every ports                       || class:port
every 443 ports                || class:port port:443
ports ending with 443      || class:port port:.*443
all 80 or 443 ports            || class:port port|80|443
every port excepted 443  || port:(?!443).* class:port
ports excepted 80|443     || port:(?!443|80).* class:port
all defects                        || class:defect
every defect with Foo      || class:defect title:.*foo.*
all domains                      || class:ip ip:.*[A-Za-z].*
every ip starting as 1.2    || ip:1\.2\..* class:ip"""
    @classmethod
    def getKeywordsSuggestion(cls, coll, start, terms):
        liste = [keyword for keyword in cls.classes[coll]["keywords"]
                 if keyword.startswith(start) and keyword not in terms]
        if coll in ["ip", "port"] and start.startswith("infos."):
            apiclient = APIClient.getInstance()
            infosInDb = apiclient.find(
                Search.classes[coll]["collection"], {"infos": {"$ne": {}}})
            keys = set()
            for infoInDb in infosInDb:
                for k in infoInDb["infos"].keys():
                    keys.add("infos."+k)
            liste = list(keys)
            liste.sort()

        return liste

    @classmethod
    def getSuggestions(cls, searchValue, entry):
        initialValue = str(searchValue.get())
        pos = entry.index('insert')
        wordPos = initialValue.find(" ", pos)
        wordBeforePos = initialValue.find(" ", 0, pos)
        nextValuePos = initialValue.find(":", pos)
        editingTheKw = wordPos <= nextValuePos
        toComplete = initialValue[:wordPos] if wordPos != -1 else initialValue
        words = toComplete.split()
        # EMPTY SEARCH BAR, return class choice
        if len(words) == 0:
            ret = list(map(lambda x: "class:"+x, cls.classes.keys()))
            return ret
        ret = []
        # get word to complete, if a space just has been placed, we have a new word
        if toComplete[wordPos] == " ":
            word = ""
        else:
            word = words[-1].strip()
        posValue = word.find(':')  # Search a key:value delimiter
        try:
            # Check if a class has been found
            coll, terms = cls.parse(toComplete, True)
        except ParseError:
            return ret
        # No class found, suggests class
        if coll == "":
            try:
                value = word[posValue+1:]
                if value != "".strip() and not editingTheKw and posValue != -1:
                    classes = [classe for classe in cls.classes.keys()
                               if classe.startswith(value)]
                else:
                    classes = cls.classes.keys()
            except IndexError:
                classes = cls.classes.keys()
            ret = list(map(lambda x: "class:"+x, classes))
            return ret
        # No : found, suggests keyword
        if posValue == -1:
            ret = list(map(lambda x: initialValue[:wordBeforePos]+" "+x, cls.getKeywordsSuggestion(
                coll, word, list(map(lambda x: x.op, terms)))))
        return ret

    def __init__(self, query="", exactMatch=True):
        try:
            self.query = query
            self.coll = ""
            Search.exact_match = exactMatch
            self.terms = []
            if self.query != "":
                self.coll, self.terms = Search.parse(query)
        except ParseError as e:
            raise e

    def getViews(self, appTw, viewFrame, mainApp):
        mongoLine = {}
        apiclient = APIClient.getInstance()
        for term in self.terms:
            key, val = term.getMongo()
            mongoLine[key] = val
        found_res = apiclient.find(
            Search.classes[self.coll]["collection"], mongoLine)
        ret = []
        for found in found_res:
            view_cls = Search.classes[self.coll]["view"]
            model_cls = Search.classes[self.coll]["model"]
            ret.append(view_cls(appTw, viewFrame, mainApp, model_cls(found)))
        return ret

    def getIds(self):
        ret = []
        apiclient = APIClient.getInstance()
        if self.coll == '':
            return ret
        mongoLine = {}
        for term in self.terms:
            key, val = term.getMongo()
            mongoLine[key] = val
        found_res = apiclient.find(
            Search.classes[self.coll]["collection"], mongoLine)
        for found in found_res:
            ret.append(str(found["_id"]))
        return ret

    @classmethod
    def parse(cls, query, ignoreException=False):
        ret = []
        vals = query.split(" ")
        coll = ""
        for val in vals:
            kv = val.split(":")
            if len(kv) != 2 and not ignoreException:
                raise ParseError("Incorrect search term: '" +
                                 str(val)+"'. Parameter name or value missing.")
            operator = kv[0].strip()
            if len(kv) == 2:
                args = kv[1].strip()
            else:
                args = ""
            if operator == "class":
                if args in Search.classes.keys():
                    coll = args
                elif not ignoreException:
                    raise ParseError(
                        "invalid class. Valid classes are "+(', '.join(Search.classes.keys())))
            else:
                try:
                    op = Operator(operator, args, Search.exact_match)
                except:
                    raise ParseError(
                        "Argument "+str(operator)+":"+str(args) + " is not a valid python regex.")
                ret.append(op)
        if coll == "" and not ignoreException:
            raise ParseError("Term class required class:<value>")
        return coll, ret



"""Describe custom tkinter widgets or extended tkinter widgets"""
import tkinter as tk

class Form(object):
    """
    Describe custom tkinter widgets or extended tkinter widgets
    Form field empty, should be inherited to create a new Form.
    """

    def __init__(self, name, style=None, **kwargs):
        """
        Constructor for an empty form.

        Args:
            name: the button text.
        """
        self.name = name
        self.val = tk.StringVar()
        self.kwargs = dict()
        self.style = None

    def constructView(self, parent):
        """
        Create the view inside the parent view given
        Args:
            parent: parent tkinter container widget
        """
        # pass

    def getValue(self):
        """
        Return the form value. Required for a form.

        Returns:
            Return None
        """
        return None
    
    def bind(self, event, callback):
        binds = self.kwargs.get("binds", {})
        binds[event] = callback
        self.kwargs["binds"] = binds

    def setValue(self, newval):
        """
        Set the form value. Required for a form.
        Args:
           newval: new value to be setted
        """
        self.val.set(newval)

    def checkForm(self):
        """
        Check if this form is correctly filled.

        Returns:
            {
                "correct": True if the form is correctly filled, False otherwise.
                "msg": A message indicating what is not correctly filled.
            }
        """
        return True, ""

    def setFocus(self):
        """Defines what item should be focused inside the form widget"""
        return

    def getKw(self, key, default):
        """Read and delete the given key inside the stored kwargs.
        If key does not exist, default will be returned.
        Args:
            key: the key matching the wanted value in kwargs
            default: a default value to be returned in case the key does not exist.
        """
        if key in self.kwargs:
            v = self.kwargs[key]
            del self.kwargs[key]
            return v
        return default


"""View for port object. Handle node in treeview and present forms to user when interacted with."""

from pollenisatorgui.core.controllers.checkinstancecontroller import CheckInstanceController
from pollenisatorgui.core.models.defect import Defect
from pollenisatorgui.core.models.command import Command
from pollenisatorgui.core.views.checkinstanceview import CheckInstanceView
from pollenisatorgui.core.views.viewelement import ViewElement
from pollenisatorgui.core.views.defectview import DefectView
from pollenisatorgui.core.controllers.portcontroller import PortController
from pollenisatorgui.core.controllers.defectcontroller import DefectController
import tkinter as tk
import json
import pollenisatorgui.core.components.utils as utils


class PortView(ViewElement):
    """View for port object. Handle node in treeview and present forms to user when interacted with.
    Attributes:
        icon: icon name to show in treeview. Icon filename must be in icon directory"""

    icon = 'port.png'

    def __init__(self, appTw, appViewFrame, mainApp, controller):
        """Constructor
        Args:
            appTw: a PollenisatorTreeview instance to put this view in
            appViewFrame: an view frame to build the forms in.
            mainApp: the Application instance
            controller: a CommandController for this view.
        """
        super().__init__(appTw, appViewFrame, mainApp, controller)
        self.tool_panel = None

    def openInsertWindow(self):
        """
        Creates a tkinter form using Forms classes. This form aims to insert a new Port
        """
        modelData = self.controller.getData()
        top_panel = self.form.addFormPanel(grid=True)
        top_panel.addFormLabel("Number")
        top_panel.addFormStr("Number", r"\d+", "", column=1)
        top_panel.addFormLabel("Proto", row=1)
        top_panel.addFormCombo("Proto", ["tcp", "udp"], "tcp", row=1, column=1)
        top_panel.addFormLabel("Service", row=2)
        top_panel.addFormStr("Service", r"", "", row=2, column=1)
        top_panel.addFormLabel("Product", row=3)
        top_panel.addFormStr("Product", r"", "", row=3, column=1)
        self.form.addFormHidden("ip", modelData["ip"])
        self.completeInsertWindow()

    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to update or delete an existing Port
        """
        self.form.clear()
        modelData = self.controller.getData()
        self.image_nav = tk.PhotoImage(file=utils.getIcon("internet.png"))
        top_panel = self.form.addFormPanel(grid=True)
        top_panel.addFormLabel("IP", row=0, column=0)
        top_panel.addFormStr(
            "IP", '', modelData["ip"], None, column=1, row=0, state="readonly")
        top_panel.addFormLabel("Number", column=0, row=1)
        top_panel.addFormStr(
            "Number", '', modelData["port"], None, column=1, row=1, state="readonly")
        top_panel.addFormLabel("Proto", row=2, column=0)
        top_panel.addFormStr(
            "Proto", '', modelData["proto"], None, column=1, row=2, state="readonly")
        top_panel.addFormLabel("Service", row=3)
        top_panel.addFormStr(
            "Service", r"", modelData["service"], column=1, row=3)
        if "http" in modelData["service"]:
            top_panel.addFormButton(
                "Open in browser", self.openInBrowser, column=2, row=3, image=self.image_nav, style="icon.TButton")
        top_panel.addFormLabel("Product", row=4)
        top_panel.addFormStr("Product", r"", modelData["product"],  row=4, column=1)
        top_panel = self.form.addFormPanel()
        top_panel.addFormLabel("Notes", side="top")
        top_panel.addFormText(
            "Notes", r"", modelData["notes"], None, side="top")
        top_panel.addFormLabel("Infos", side="left")
        top_panel.addFormText("Infos", utils.is_json, json.dumps(modelData["infos"], indent=4, cls=utils.JSONEncoder), height=100, side="left", fill="both")
        # command_list = Command.fetchObjects({"lvl": "port"}, APIClient.getInstance().getCurrentPentest()) #TODO lvl change
        # command_names = ["None"]
        # self.command_names_to_iid = {}
        # for command_doc in command_list:
        #     command_names.append(command_doc.name)
        #     self.command_names_to_iid[command_doc.name] = str(command_doc._id)
        # self.tool_panel = self.form.addFormPanel(grid=True)
        # self.tool_panel.addFormLabel("Tool to add")
        # self.tool_panel.addFormCombo(
        #     "Tool to add", command_names, "None", column=1)
        # self.tool_panel.addFormButton("Add tool", self._addTool, column=2)
        top_panel = self.form.addFormPanel(grid=True)
        top_panel.addFormButton("Add a security defect",
                                self.addDefectCallback)
        self.form.addFormHidden("ip", modelData["ip"])
        self.completeModifyWindow()

    def openInBrowser(self, _event=None):
        """Callback for action open in browser
        Args:
            _event: nut used but mandatory
        """
        self.controller.model.openInBrowser()

    def getAdditionalContextualCommands(self):
        return {"Open In Browser":self.openInBrowser}

    def addDefectCallback(self, _event=None):
        """
        Create an empty defect model and its attached view. Open this view insert window.

        Args:
            event: Automatically generated with a button Callback, not used but mandatory.
        """
        for widget in self.appliViewFrame.winfo_children():
            widget.destroy()
        modelData = self.controller.getData()
        modelData["target_type"] = self.controller.model.__class__.__name__.lower()
        modelData["target_id"] = self.controller.model.getId()
        dv = DefectView(self.appliTw, self.appliViewFrame,
                        self.mainApp, DefectController(Defect(modelData)))
        dv.openInsertWindow()

    def _addTool(self, _event=None):
        """Callback for add tool action
        Add a tool without any check to port.
        Args:
            _event: not used but mandatory
        """
        toolname_values = self.tool_panel.getValue()
        toolname = ViewElement.list_tuple_to_dict(toolname_values)[
            "Tool to add"]
        command_iid = self.command_names_to_iid.get(toolname)
        if command_iid is not None:
            self.controller.addCustomTool(command_iid)

    def _insertChildren(self):
        defects = self.controller.getDefects()
        for defect in defects:
            if isinstance(defect, dict):
                defect = Defect(defect)
            defect_o = DefectController(defect)
            defect_vw = DefectView(
                self.appliTw, self.appliViewFrame, self.mainApp, defect_o)
            defect_vw.addInTreeview(str(self.controller.getDbId()), addChildren=False)

        checks = self.controller.getChecks()
        for check in checks:
            check_o = CheckInstanceController(check)
            check_vw = CheckInstanceView(self.appliTw, self.appliViewFrame, self.mainApp, check_o)
            check_vw.addInTreeview(str(self.controller.getDbId()), addChildren=False)

    @classmethod
    def multiAddInTreeview(self, appliTw, appliViewFrame, mainApp, ports, parent, **kwargs):
        for port in ports:
            port_o = PortController(port)
            port_vw = PortView(
                appliTw, appliViewFrame, mainApp, port_o)
            port_vw.addInTreeview(parent, addChildren=False, detailed=kwargs.get("detailed", False))
        if kwargs.get("addChildren", True):
            checks = PortController.getChecksForPorts(ports)
            defects = PortController.getDefectsForPorts(ports)
            for check in checks:
                check_o = CheckInstanceController(check)
                check_vw = CheckInstanceView(appliTw, appliViewFrame, mainApp, check_o)
                check_vw.addInTreeview(None, addChildren=False)
            for defect in defects:
                defect_o = DefectController(Defect(defect))
                defect_vw = DefectView(
                    appliTw, appliViewFrame, mainApp, defect_o)
                defect_vw.addInTreeview(None, addChildren=False)
                
    def addInTreeview(self, parentNode=None, **kwargs):
        """Add this view in treeview. Also stores infos in application treeview.
        Args:
            parentNode: if None, will calculate the parent. If setted, forces the node to be inserted inside given parentNode.
            addChildren: If False, skip the tool and defects insert. Useful when displaying search results
        """
        addChildren = kwargs.get("addChildren", True)
        if kwargs.get("detailed", False):
            nodeText = self.controller.getDetailedString()
        elif parentNode is None:
            parentNode = self.getParentNode()
            nodeText = str(self.controller.getModelRepr())
        elif parentNode == '':
            nodeText = self.controller.getDetailedString()
        else:
            nodeText = str(self.controller.getModelRepr())
        self.appliTw.views[str(self.controller.getDbId())] = {"view": self}
        tags = self.controller.getTags()
        try:
             #CALLING TK CALL IS FASTER than
            # self.appliTw.insert(parentNode, "end", str(
            #     self.controller.getDbId()), text=nodeText, tags=self.controller.getTags(), image=self.getClassIcon())
            port_node = self.appliTw.tk.call(self.appliTw._w, "insert", parentNode, "end", "-id", str(self.controller.getDbId()), 
                                 "-text", str(nodeText), "-tags", tags, "-image", self.getClassIcon())
        
        except tk.TclError:
            pass
        if addChildren:
            self._insertChildren()
        if self.mainApp.settings.is_checklist_view():
            self.hide("checklist_view")
        if "hidden" in tags:
            self.hide("tags")

    def key(self):
        """Returns a key for sorting this node
        Returns:
            Tuple of 1 integer valus representing the prot number
            
        """
        return tuple([int(self.controller.getData()["port"])])

    def updateReceived(self, obj=None, old_obj=None):
        """Called when a port update is received by notification.
        Update the port node in summary
        """
        for module in self.mainApp.modules:
            if callable(getattr(module["object"], "updatePort", None)):
                module["object"].updatePort(self.controller.getData())
        super().updateReceived()

    def insertReceived(self):
        """Called when a port insertion is received by notification.
        Insert the node in summary.
        """
        for module in self.mainApp.modules:
            if callable(getattr(module["object"], "insertPort", None)):
                module["object"].insertPort(self.controller.getData())



from bson import ObjectId
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.models.element import Element
from pollenisatorgui.modules.ActiveDirectory.models.computer_infos import ComputerInfos

class Computer(Element):
    coll_name = "computers"
    classes = ["computer"]

    @property
    def infos(self):
        """Gets the infos of this Computer.


        :return: The infos of this Computer.
        :rtype: ComputerInfos
        """
        return self._infos
    
    @infos.setter
    def infos(self, infos):
        """Sets the infos of this Computer.


        :param infos: The infos of this Computer.
        :type infos: ComputerInfos
        """
        #keeping clarity with explicit checks
        self._infos = infos
    
    def __init__(self, valuesFromDb=None):
        if valuesFromDb is None:
            valuesFromDb = {}
        super().__init__(valuesFromDb.get("_id", None), valuesFromDb.get("parent", None),  valuesFromDb.get("infos", {}))
        self.initialize(valuesFromDb.get("_id"),  valuesFromDb.get("name"), valuesFromDb.get("ip"), \
             valuesFromDb.get("domain"),  valuesFromDb.get("admins"),  valuesFromDb.get("users"), valuesFromDb.get("infos"))

    def initialize(self, _id=None, name=None, ip=None, domain=None, admins=None, users=None, infos=None):  # noqa: E501
        """Computer - a model defined in OpenAPI
        :param _id: iid of the object
        :type _id: str
        :param name: The name of this Computer.  # noqa: E501
        :type name: str
        :param ip: The ip of this Computer.  # noqa: E501
        :type ip: str
        :param domain: The domain of this Computer.  # noqa: E501
        :type domain: str
        :param admins: The admins of this Computer.  # noqa: E501
        :type admins: List[str]
        :param users: The users of this Computer.  # noqa: E501
        :type users: List[str]
        :param infos: The infos of this Computer.  # noqa: E501
        :type infos: ComputerInfos
        """

        self._id = _id
        self.name = name
        self.ip = ip
        self.domain = domain
        self.admins = admins
        self.users = users
        self._infos = ComputerInfos(infos)
    
    def update(self, pipeline_set):
        """Update this object in database.
        Args:
            pipeline_set: (Opt.) A dictionnary with custom values. If None (default) use model attributes.
        """
        apiclient = APIClient.getInstance()
        if pipeline_set is None:
            raise NotImplementedError("Computer.update() not implemented")
        else:
            apiclient.update("computers", ObjectId(self._id), pipeline_set)

    def __str__(self):
        """
        Get a string representation of a defect.

        Returns:
            Returns the defect +title.
        """
        return str(self.domain)+"\\"+str(self.name) + " ("+str(self.ip)+")"

    def getData(self):
        return {"_id": self._id, "name":self.name, "ip":self.ip, "domain":self.domain,
            "admins":self.admins, "users": self.users, "infos":self.infos.getData()}

    def _getParentId(self):
        """
        Return the mongo ObjectId _id of the first parent of this object. For a port it is the ip.

        Returns:
            Returns the parent ip's ObjectId _id".
        """
        datamanager = DataManager.getInstance()
        obj = datamanager.find("ips", {"ip":self.ip}, False)
        if obj is None:
            return None
        return obj.getId()
    
    @classmethod
    def fetchObjects(cls, pipeline):
        """Fetch many commands from database and return a Cursor to iterate over model objects
        Args:
            pipeline: a Mongo search pipeline (dict)
        Returns:
            Returns a cursor to iterate on model objects
        """
        apiclient = APIClient.getInstance()
        pipeline["type"] = "computer"
        ds = apiclient.find(cls.coll_name, pipeline, True)
        if ds is None:
            return None
        for d in ds:
            # disabling this error as it is an abstract function
            yield cls(d)  #  pylint: disable=no-value-for-parameter
    
    @classmethod
    def fetchObject(cls, pipeline):
        """Fetch many commands from database and return a Cursor to iterate over model objects
        Args:
            pipeline: a Mongo search pipeline (dict)
        Returns:
            Returns a cursor to iterate on model objects
        """
        apiclient = APIClient.getInstance()
        pipeline["type"] = "computer"
        d = apiclient.find(cls.coll_name, pipeline, False)
        if d is None:
            return None
        return cls(d)
    
    @classmethod
    def fetchPentestObjects(cls):
        return [x for x in cls.fetchObjects({})]